@@ -29,28 +29,27 @@
 #include <fcntl.h>
 #include <limits.h>
 
-#include <event.h>
-
 #include <libavcodec/avcodec.h>
 #include <libavformat/avformat.h>
 #include <libswscale/swscale.h>
 
 #include "db.h"
+#include "misc.h"
 #include "logger.h"
+#include "conffile.h"
+
 #if LIBAVFORMAT_VERSION_MAJOR >= 53
 # include "avio_evbuffer.h"
 #endif
 #include "artwork.h"
 
-
-static const char *cover_basename[] =
-  {
-    "artwork", "cover",
-  };
+#ifdef HAVE_SPOTIFY_H
+# include "spotify.h"
+#endif
 
 static const char *cover_extension[] =
   {
-    "png", "jpg",
+    "jpg", "png",
   };
 
 
@@ -98,6 +97,56 @@ artwork_read(char *filename, struct evbuffer *evbuf)
   return -1;
 }
 
+static int
+rescale_needed(AVCodecContext *src, int max_w, int max_h, int *target_w, int *target_h)
+{
+  int need_rescale;
+
+  DPRINTF(E_DBG, L_ART, "Original image dimensions: w %d h %d\n", src->width, src->height);
+
+  need_rescale = 1;
+
+  if ((max_w <= 0) || (max_h <= 0)) /* No valid dimensions, use original */
+    {
+      need_rescale = 0;
+
+      *target_w = src->width;
+      *target_h = src->height;
+    }
+  else if ((src->width <= max_w) && (src->height <= max_h)) /* Smaller than target */
+    {
+      need_rescale = 0;
+
+      *target_w = src->width;
+      *target_h = src->height;
+    }
+  else if (src->width * max_h > src->height * max_w)   /* Wider aspect ratio than target */
+    {
+      *target_w = max_w;
+      *target_h = (double)max_w * ((double)src->height / (double)src->width);
+    }
+  else                                                 /* Taller or equal aspect ratio */
+    {
+      *target_w = (double)max_h * ((double)src->width / (double)src->height);
+      *target_h = max_h;
+    }
+
+  DPRINTF(E_DBG, L_ART, "Raw destination width %d height %d\n", *target_w, *target_h);
+
+  if ((*target_h > max_h) && (max_h > 0))
+    *target_h = max_h;
+
+  /* PNG prefers even row count */
+  *target_w += *target_w % 2;
+
+  if ((*target_w > max_w) && (max_w > 0))
+    *target_w = max_w - (max_w % 2);
+
+  DPRINTF(E_DBG, L_ART, "Destination width %d height %d\n", *target_w, *target_h);
+
+  return need_rescale;
+}
+
 static int
 artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int format, struct evbuffer *evbuf)
 {
@@ -114,9 +163,6 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
   AVCodec *img_decoder;
   AVCodec *img_encoder;
 
-  int64_t pix_fmt_mask;
-  const enum PixelFormat *pix_fmts;
-
   AVFrame *i_frame;
   AVFrame *o_frame;
 
@@ -139,7 +185,11 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
       return -1;
     }
 
+#if LIBAVCODEC_VERSION_MAJOR >= 54 || (LIBAVCODEC_VERSION_MAJOR == 53 && LIBAVCODEC_VERSION_MINOR >= 6)
+  ret = avcodec_open2(src, img_decoder, NULL);
+#else
   ret = avcodec_open(src, img_decoder);
+#endif
   if (ret < 0)
     {
       DPRINTF(E_LOG, L_ART, "Could not open codec for decoding: %s\n", strerror(AVUNERROR(ret)));
@@ -162,6 +212,24 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
       goto out_close_src;
     }
 
+#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
+  dst_fmt->video_codec = AV_CODEC_ID_NONE;
+
+  /* Try to keep same codec if possible */
+  if ((src->codec_id == AV_CODEC_ID_PNG) && (format & ART_CAN_PNG))
+    dst_fmt->video_codec = AV_CODEC_ID_PNG;
+  else if ((src->codec_id == AV_CODEC_ID_MJPEG) && (format & ART_CAN_JPEG))
+    dst_fmt->video_codec = AV_CODEC_ID_MJPEG;
+
+  /* If not possible, select new codec */
+  if (dst_fmt->video_codec == AV_CODEC_ID_NONE)
+    {
+      if (format & ART_CAN_PNG)
+	dst_fmt->video_codec = AV_CODEC_ID_PNG;
+      else if (format & ART_CAN_JPEG)
+	dst_fmt->video_codec = AV_CODEC_ID_MJPEG;
+    }
+#else
   dst_fmt->video_codec = CODEC_ID_NONE;
 
   /* Try to keep same codec if possible */
@@ -178,6 +246,7 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
       else if (format & ART_CAN_JPEG)
 	dst_fmt->video_codec = CODEC_ID_MJPEG;
     }
+#endif
 
   img_encoder = avcodec_find_encoder(dst_fmt->video_codec);
   if (!img_encoder)
@@ -212,7 +281,11 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
     }
 #endif
 
+#if LIBAVFORMAT_VERSION_MAJOR >= 54 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVFORMAT_VERSION_MINOR >= 21)
+  dst_st = avformat_new_stream(dst_ctx, NULL);
+#else
   dst_st = av_new_stream(dst_ctx, 0);
+#endif
   if (!dst_st)
     {
       DPRINTF(E_LOG, L_ART, "Out of memory for new output stream\n");
@@ -223,10 +296,10 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
 
   dst = dst_st->codec;
 
-#if LIBAVCODEC_VERSION_MAJOR >= 53 || (LIBAVCODEC_VERSION_MAJOR == 52 && LIBAVCODEC_VERSION_MINOR >= 64)
-  avcodec_get_context_defaults2(dst, AVMEDIA_TYPE_VIDEO);
+#if LIBAVCODEC_VERSION_MAJOR >= 54 || (LIBAVCODEC_VERSION_MAJOR == 53 && LIBAVCODEC_VERSION_MINOR >= 35)
+  avcodec_get_context_defaults3(dst, NULL);
 #else
-  avcodec_get_context_defaults2(dst, CODEC_TYPE_VIDEO);
+  avcodec_get_context_defaults2(dst, AVMEDIA_TYPE_VIDEO);
 #endif
 
   if (dst_fmt->flags & AVFMT_GLOBALHEADER)
@@ -239,7 +312,16 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
   dst->codec_type = CODEC_TYPE_VIDEO;
 #endif
 
-  pix_fmt_mask = 0;
+#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
+# ifndef FFMPEG_INCOMPATIBLE_API
+  dst->pix_fmt = avcodec_find_best_pix_fmt2((enum AVPixelFormat *)img_encoder->pix_fmts, src->pix_fmt, 1, NULL);
+# else
+  dst->pix_fmt = avcodec_find_best_pix_fmt_of_list((enum AVPixelFormat *)img_encoder->pix_fmts, src->pix_fmt, 1, NULL);
+# endif
+#else
+  const enum PixelFormat *pix_fmts;
+  int64_t pix_fmt_mask = 0;
+
   pix_fmts = img_encoder->pix_fmts;
   while (pix_fmts && (*pix_fmts != -1))
     {
@@ -248,13 +330,14 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
     }
 
   dst->pix_fmt = avcodec_find_best_pix_fmt(pix_fmt_mask, src->pix_fmt, 1, NULL);
+#endif
 
   if (dst->pix_fmt < 0)
     {
       DPRINTF(E_LOG, L_ART, "Could not determine best pixel format\n");
 
       ret = -1;
-      goto out_free_dst;
+      goto out_free_dst_ctx;
     }
 
   DPRINTF(E_DBG, L_ART, "Selected pixel format: %d\n", dst->pix_fmt);
@@ -265,29 +348,38 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
   dst->width = out_w;
   dst->height = out_h;
 
-#if LIBAVFORMAT_VERSION_MAJOR <= 52 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVCODEC_VERSION_MINOR <= 1)
+#if LIBAVFORMAT_VERSION_MAJOR <= 52 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVFORMAT_VERSION_MINOR <= 1)
   ret = av_set_parameters(dst_ctx, NULL);
   if (ret < 0)
     {
       DPRINTF(E_LOG, L_ART, "Invalid parameters for artwork output: %s\n", strerror(AVUNERROR(ret)));
 
       ret = -1;
-      goto out_free_dst;
+      goto out_free_dst_ctx;
     }
 #endif
 
   /* Open encoder */
+#if LIBAVCODEC_VERSION_MAJOR >= 54 || (LIBAVCODEC_VERSION_MAJOR == 53 && LIBAVCODEC_VERSION_MINOR >= 6)
+  ret = avcodec_open2(dst, img_encoder, NULL);
+#else
   ret = avcodec_open(dst, img_encoder);
+#endif
   if (ret < 0)
     {
       DPRINTF(E_LOG, L_ART, "Could not open codec for encoding: %s\n", strerror(AVUNERROR(ret)));
 
       ret = -1;
-      goto out_free_dst;
+      goto out_free_dst_ctx;
     }
 
+#if LIBAVCODEC_VERSION_MAJOR >= 56 || (LIBAVCODEC_VERSION_MAJOR == 55 && LIBAVCODEC_VERSION_MINOR >= 29)
+  i_frame = av_frame_alloc();
+  o_frame = av_frame_alloc();
+#else
   i_frame = avcodec_alloc_frame();
   o_frame = avcodec_alloc_frame();
+#endif
 
   if (!i_frame || !o_frame)
     {
@@ -399,6 +491,24 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
       goto out_free_buf;
     }
 
+#if LIBAVCODEC_VERSION_MAJOR >= 54
+  av_init_packet(&pkt);
+  pkt.data = outbuf;
+  pkt.size = outbuf_len;
+  ret = avcodec_encode_video2(dst, &pkt, o_frame, &have_frame);
+  if (!ret && have_frame && dst->coded_frame) 
+    {
+      dst->coded_frame->pts       = pkt.pts;
+      dst->coded_frame->key_frame = !!(pkt.flags & AV_PKT_FLAG_KEY);
+    }
+  else if (ret < 0)
+    {
+      DPRINTF(E_LOG, L_ART, "Could not encode artwork\n");
+
+      ret = -1;
+      goto out_fclose_dst;
+    }
+#else
   ret = avcodec_encode_video(dst, outbuf, outbuf_len, o_frame);
   if (ret <= 0)
     {
@@ -412,8 +522,9 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
   pkt.stream_index = 0;
   pkt.data = outbuf;
   pkt.size = ret;
+#endif
 
-#if LIBAVFORMAT_VERSION_MAJOR >= 53 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVCODEC_VERSION_MINOR >= 3)
+#if LIBAVFORMAT_VERSION_MAJOR >= 54 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVFORMAT_VERSION_MINOR >= 3)
   ret = avformat_write_header(dst_ctx, NULL);
 #else
   ret = av_write_header(dst_ctx);
@@ -447,11 +558,19 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
 
   switch (dst_fmt->video_codec)
     {
+#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
+      case AV_CODEC_ID_PNG:
+#else
       case CODEC_ID_PNG:
+#endif
 	ret = ART_FMT_PNG;
 	break;
 
+#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
+      case AV_CODEC_ID_MJPEG:
+#else
       case CODEC_ID_MJPEG:
+#endif
 	ret = ART_FMT_JPEG;
 	break;
 
@@ -473,18 +592,26 @@ artwork_rescale(AVFormatContext *src_ctx, int s, int out_w, int out_h, int forma
   av_free(buf);
 
  out_free_frames:
+#if LIBAVCODEC_VERSION_MAJOR >= 56 || (LIBAVCODEC_VERSION_MAJOR == 55 && LIBAVCODEC_VERSION_MINOR >= 29)
+  if (i_frame)
+    av_frame_free(&i_frame);
+  if (o_frame)
+    av_frame_free(&o_frame);
+#elif LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
+  if (i_frame)
+    avcodec_free_frame(&i_frame);
+  if (o_frame)
+    avcodec_free_frame(&o_frame);
+#else
   if (i_frame)
     av_free(i_frame);
   if (o_frame)
     av_free(o_frame);
+#endif
   avcodec_close(dst);
 
- out_free_dst:
-  av_free(dst_st);
-  av_free(dst);
-
  out_free_dst_ctx:
-  av_free(dst_ctx);
+  avformat_free_context(dst_ctx);
 
  out_close_src:
   avcodec_close(src);
@@ -496,19 +623,17 @@ static int
 artwork_get(char *filename, int max_w, int max_h, int format, struct evbuffer *evbuf)
 {
   AVFormatContext *src_ctx;
-  AVCodecContext *src;
   int s;
   int target_w;
   int target_h;
-  int need_rescale;
   int format_ok;
   int ret;
 
-  DPRINTF(E_DBG, L_ART, "Artwork request parameters: max w = %d, max h = %d\n", max_w, max_h);
+  DPRINTF(E_DBG, L_ART, "Getting artwork (max destination width %d height %d)\n", max_w, max_h);
 
   src_ctx = NULL;
 
-#if LIBAVFORMAT_VERSION_MAJOR >= 53 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVCODEC_VERSION_MINOR >= 3)
+#if LIBAVFORMAT_VERSION_MAJOR >= 54 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVFORMAT_VERSION_MINOR >= 3)
   ret = avformat_open_input(&src_ctx, filename, NULL, NULL);
 #else
   ret = av_open_input_file(&src_ctx, filename, NULL, 0, NULL);
@@ -520,24 +645,40 @@ artwork_get(char *filename, int max_w, int max_h, int format, struct evbuffer *e
       return -1;
     }
 
+#if LIBAVFORMAT_VERSION_MAJOR >= 54 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVFORMAT_VERSION_MINOR >= 3)
+  ret = avformat_find_stream_info(src_ctx, NULL);
+#else
   ret = av_find_stream_info(src_ctx);
+#endif
   if (ret < 0)
     {
       DPRINTF(E_WARN, L_ART, "Cannot get stream info: %s\n", strerror(AVUNERROR(ret)));
 
+#if LIBAVFORMAT_VERSION_MAJOR >= 54 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVFORMAT_VERSION_MINOR >= 21)
+      avformat_close_input(&src_ctx);
+#else
       av_close_input_file(src_ctx);
+#endif
       return -1;
     }
 
   format_ok = 0;
   for (s = 0; s < src_ctx->nb_streams; s++)
     {
+#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
+      if (src_ctx->streams[s]->codec->codec_id == AV_CODEC_ID_PNG)
+#else
       if (src_ctx->streams[s]->codec->codec_id == CODEC_ID_PNG)
+#endif
 	{
 	  format_ok = (format & ART_CAN_PNG) ? ART_FMT_PNG : 0;
 	  break;
 	}
+#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
+      else if (src_ctx->streams[s]->codec->codec_id == AV_CODEC_ID_MJPEG)
+#else
       else if (src_ctx->streams[s]->codec->codec_id == CODEC_ID_MJPEG)
+#endif
 	{
 	  format_ok = (format & ART_CAN_JPEG) ? ART_FMT_JPEG : 0;
 	  break;
@@ -548,55 +689,18 @@ artwork_get(char *filename, int max_w, int max_h, int format, struct evbuffer *e
     {
       DPRINTF(E_LOG, L_ART, "Artwork file '%s' not a PNG or JPEG file\n", filename);
 
+#if LIBAVFORMAT_VERSION_MAJOR >= 54 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVFORMAT_VERSION_MINOR >= 21)
+      avformat_close_input(&src_ctx);
+#else
       av_close_input_file(src_ctx);
+#endif
       return -1;
     }
 
-  src = src_ctx->streams[s]->codec;
-
-  DPRINTF(E_DBG, L_ART, "Original image '%s': w %d h %d\n", filename, src->width, src->height);
-
-  need_rescale = 1;
-
-  /* Determine width/height -- assuming max_w == max_h */
-  if ((src->width <= max_w) && (src->height <= max_h))
-    {
-      need_rescale = 0;
-
-      target_w = src->width;
-      target_h = src->height;
-    }
-  else if (src->width > src->height)
-    {
-      target_w = max_w;
-      target_h = (double)max_h * ((double)src->height / (double)src->width);
-    }
-  else if (src->height > src->width)
-    {
-      target_h = max_h;
-      target_w = (double)max_w * ((double)src->width / (double)src->height);
-    }
-  else
-    {
-      target_w = max_w;
-      target_h = max_h;
-    }
-
-  DPRINTF(E_DBG, L_ART, "Raw destination width %d height %d\n", target_w, target_h);
-
-  if (target_h > max_h)
-    target_h = max_h;
-
-  /* PNG prefers even row count */
-  target_w += target_w % 2;
-
-  if (target_w > max_w)
-    target_w = max_w - (max_w % 2);
-
-  DPRINTF(E_DBG, L_ART, "Destination width %d height %d\n", target_w, target_h);
+  ret = rescale_needed(src_ctx->streams[s]->codec, max_w, max_h, &target_w, &target_h);
 
   /* Fastpath */
-  if (!need_rescale && format_ok)
+  if (!ret && format_ok)
     {
       ret = artwork_read(filename, evbuf);
       if (ret == 0)
@@ -605,7 +709,11 @@ artwork_get(char *filename, int max_w, int max_h, int format, struct evbuffer *e
   else
     ret = artwork_rescale(src_ctx, s, target_w, target_h, format, evbuf);
 
+#if LIBAVFORMAT_VERSION_MAJOR >= 54 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVFORMAT_VERSION_MINOR >= 21)
+  avformat_close_input(&src_ctx);
+#else
   av_close_input_file(src_ctx);
+#endif
 
   if (ret < 0)
     {
@@ -616,6 +724,127 @@ artwork_get(char *filename, int max_w, int max_h, int format, struct evbuffer *e
   return ret;
 }
 
+#if LIBAVFORMAT_VERSION_MAJOR >= 55 || (LIBAVFORMAT_VERSION_MAJOR == 54 && LIBAVFORMAT_VERSION_MINOR >= 6)
+static int
+artwork_get_embedded_image(char *filename, int max_w, int max_h, int format, struct evbuffer *evbuf)
+{
+  AVFormatContext *src_ctx;
+  AVStream *src_st;
+  int s;
+  int target_w;
+  int target_h;
+  int format_ok;
+  int ret;
+
+  /* If item is an internet stream don't look for artwork */
+  if (strncmp(filename, "http://", strlen("http://")) == 0)
+    return -1;
+
+  /* If Spotify item don't look for artwork */
+  if (strncmp(filename, "spotify:", strlen("spotify:")) == 0)
+    return -1;
+
+  DPRINTF(E_SPAM, L_ART, "Trying embedded artwork in %s\n", filename);
+
+  src_ctx = NULL;
+
+  ret = avformat_open_input(&src_ctx, filename, NULL, NULL);
+  if (ret < 0)
+    {
+      DPRINTF(E_WARN, L_ART, "Cannot open media file '%s': %s\n", filename, strerror(AVUNERROR(ret)));
+
+      return -1;
+    }
+
+  ret = avformat_find_stream_info(src_ctx, NULL);
+  if (ret < 0)
+    {
+      DPRINTF(E_WARN, L_ART, "Cannot get stream info: %s\n", strerror(AVUNERROR(ret)));
+
+      avformat_close_input(&src_ctx);
+      return -1;
+    }
+
+  format_ok = 0;
+  for (s = 0; s < src_ctx->nb_streams; s++)
+    {
+      if (src_ctx->streams[s]->disposition & AV_DISPOSITION_ATTACHED_PIC)
+	{
+#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
+	  if (src_ctx->streams[s]->codec->codec_id == AV_CODEC_ID_PNG)
+#else
+	  if (src_ctx->streams[s]->codec->codec_id == CODEC_ID_PNG)
+#endif
+	    {
+	      format_ok = (format & ART_CAN_PNG) ? ART_FMT_PNG : 0;
+	      break;
+	    }
+#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
+	  else if (src_ctx->streams[s]->codec->codec_id == AV_CODEC_ID_MJPEG)
+#else
+	  else if (src_ctx->streams[s]->codec->codec_id == CODEC_ID_MJPEG)
+#endif
+	    {
+	      format_ok = (format & ART_CAN_JPEG) ? ART_FMT_JPEG : 0;
+	      break;
+	    }
+	}
+    }
+
+  if (s == src_ctx->nb_streams)
+    {
+      DPRINTF(E_SPAM, L_ART, "Did not find embedded artwork in '%s'\n", filename);
+
+      avformat_close_input(&src_ctx);
+      return -1;
+    }
+  else
+    DPRINTF(E_DBG, L_ART, "Found embedded artwork in '%s'\n", filename);
+
+  src_st = src_ctx->streams[s];
+
+  ret = rescale_needed(src_st->codec, max_w, max_h, &target_w, &target_h);
+
+  /* Fastpath */
+  if (!ret && format_ok)
+    {
+      DPRINTF(E_DBG, L_ART, "Artwork not too large, using original image\n");
+
+      ret = evbuffer_expand(evbuf, src_st->attached_pic.size);
+      if (ret < 0)
+	{
+	  DPRINTF(E_LOG, L_ART, "Out of memory for artwork\n");
+
+	  avformat_close_input(&src_ctx);
+	  return -1;
+	}
+
+      ret = evbuffer_add(evbuf, src_st->attached_pic.data, src_st->attached_pic.size);
+      if (ret < 0)
+	{
+	  DPRINTF(E_LOG, L_ART, "Could not add embedded image to event buffer\n");
+	}
+      else
+        ret = format_ok;
+    }
+  else
+    {
+      DPRINTF(E_DBG, L_ART, "Artwork too large, rescaling image\n");
+
+      ret = artwork_rescale(src_ctx, s, target_w, target_h, format, evbuf);
+    }
+
+  avformat_close_input(&src_ctx);
+
+  if (ret < 0)
+    {
+      if (EVBUFFER_LENGTH(evbuf) > 0)
+	evbuffer_drain(evbuf, EVBUFFER_LENGTH(evbuf));
+    }
+
+  return ret;
+}
+#endif
 
 static int
 artwork_get_own_image(char *path, int max_w, int max_h, int format, struct evbuffer *evbuf)
@@ -626,6 +855,27 @@ artwork_get_own_image(char *path, int max_w, int max_h, int format, struct evbuf
   int i;
   int ret;
 
+  /* If item is an internet stream don't look for artwork */
+  if (strncmp(path, "http://", strlen("http://")) == 0)
+    return -1;
+
+  if (strncmp(path, "spotify:", strlen("spotify:")) == 0)
+#ifdef HAVE_SPOTIFY_H
+    {
+      if (!(format & ART_CAN_JPEG))
+	return -1;
+
+      ret = spotify_artwork_get(evbuf, path, max_w, max_h);
+
+      if (ret < 0)
+	return -1;
+      else
+        return ART_FMT_JPEG;
+    }
+#else
+    return -1;
+#endif
+
   ret = snprintf(artwork, sizeof(artwork), "%s", path);
   if ((ret < 0) || (ret >= sizeof(artwork)))
     {
@@ -650,7 +900,7 @@ artwork_get_own_image(char *path, int max_w, int max_h, int format, struct evbuf
 	  continue;
 	}
 
-      DPRINTF(E_DBG, L_ART, "Trying own artwork file %s\n", artwork);
+      DPRINTF(E_SPAM, L_ART, "Trying own artwork file %s\n", artwork);
 
       ret = access(artwork, F_OK);
       if (ret < 0)
@@ -662,6 +912,8 @@ artwork_get_own_image(char *path, int max_w, int max_h, int format, struct evbuf
   if (i == (sizeof(cover_extension) / sizeof(cover_extension[0])))
     return -1;
 
+  DPRINTF(E_DBG, L_ART, "Found own artwork file %s\n", artwork);
+
   return artwork_get(artwork, max_w, max_h, format, evbuf);
 }
 
@@ -674,6 +926,16 @@ artwork_get_dir_image(char *path, int isdir, int max_w, int max_h, int format, s
   int j;
   int len;
   int ret;
+  cfg_t *lib;
+  int nbasenames;
+
+  /* If item is an internet stream don't look for artwork */
+  if (strncmp(path, "http://", strlen("http://")) == 0)
+    return -1;
+
+  /* If Spotify item don't look for artwork */
+  if (strncmp(path, "spotify:", strlen("spotify:")) == 0)
+    return -1;
 
   ret = snprintf(artwork, sizeof(artwork), "%s", path);
   if ((ret < 0) || (ret >= sizeof(artwork)))
@@ -692,19 +954,25 @@ artwork_get_dir_image(char *path, int isdir, int max_w, int max_h, int format, s
 
   len = strlen(artwork);
 
-  for (i = 0; i < (sizeof(cover_basename) / sizeof(cover_basename[0])); i++)
+  lib = cfg_getsec(cfg, "library");
+  nbasenames = cfg_size(lib, "artwork_basenames");
+
+  if (nbasenames == 0)
+    return -1;
+
+  for (i = 0; i < nbasenames; i++)
     {
       for (j = 0; j < (sizeof(cover_extension) / sizeof(cover_extension[0])); j++)
 	{
-	  ret = snprintf(artwork + len, sizeof(artwork) - len, "/%s.%s", cover_basename[i], cover_extension[j]);
+	  ret = snprintf(artwork + len, sizeof(artwork) - len, "/%s.%s", cfg_getnstr(lib, "artwork_basenames", i), cover_extension[j]);
 	  if ((ret < 0) || (ret >= sizeof(artwork) - len))
 	    {
-	      DPRINTF(E_INFO, L_ART, "Artwork path exceeds PATH_MAX (%s.%s)\n", cover_basename[i], cover_extension[j]);
+	      DPRINTF(E_INFO, L_ART, "Artwork path exceeds PATH_MAX (%s.%s)\n", cfg_getnstr(lib, "artwork_basenames", i), cover_extension[j]);
 
 	      continue;
 	    }
 
-	  DPRINTF(E_DBG, L_ART, "Trying directory artwork file %s\n", artwork);
+	  DPRINTF(E_SPAM, L_ART, "Trying directory artwork file %s\n", artwork);
 
 	  ret = access(artwork, F_OK);
 	  if (ret < 0)
@@ -717,19 +985,94 @@ artwork_get_dir_image(char *path, int isdir, int max_w, int max_h, int format, s
 	break;
     }
 
-  if (i == (sizeof(cover_basename) / sizeof(cover_basename[0])))
+  if (i == nbasenames)
     return -1;
 
+  DPRINTF(E_DBG, L_ART, "Found directory artwork file %s\n", artwork);
+
   return artwork_get(artwork, max_w, max_h, format, evbuf);
 }
 
+static int
+artwork_get_parentdir_image(char *path, int isdir, int max_w, int max_h, int format, struct evbuffer *evbuf)
+{
+  char artwork[PATH_MAX];
+  char parentdir[PATH_MAX];
+  char *ptr;
+  int len;
+  int i;
+  int ret;
+
+  /* If item is an internet stream don't look for artwork */
+  if (strncmp(path, "http://", strlen("http://")) == 0)
+    return -1;
+
+  /* If Spotify item don't look for artwork */
+  if (strncmp(path, "spotify:", strlen("spotify:")) == 0)
+    return -1;
+
+  ret = snprintf(artwork, sizeof(artwork), "%s", path);
+  if ((ret < 0) || (ret >= sizeof(artwork)))
+    {
+      DPRINTF(E_INFO, L_ART, "Artwork path exceeds PATH_MAX\n");
+
+      return -1;
+    }
+
+  if (!isdir)
+    {
+      ptr = strrchr(artwork, '/');
+      if (ptr)
+	*ptr = '\0';
+    }
+
+  ptr = strrchr(artwork, '/');
+  if ((!ptr) || (strlen(ptr) <= 1))
+    return -1;
+  strcpy(parentdir, ptr + 1);
+
+  len = strlen(artwork);
+
+  for (i = 0; i < (sizeof(cover_extension) / sizeof(cover_extension[0])); i++)
+    {
+      ret = snprintf(artwork + len, sizeof(artwork) - len, "/%s.%s", parentdir, cover_extension[i]);
+      if ((ret < 0) || (ret >= sizeof(artwork) - len))
+	{
+	  DPRINTF(E_INFO, L_ART, "Artwork path exceeds PATH_MAX (%s.%s)\n", parentdir, cover_extension[i]);
+
+	  continue;
+	}
+
+      DPRINTF(E_SPAM, L_ART, "Trying parent directory artwork file %s\n", artwork);
+
+      ret = access(artwork, F_OK);
+      if (ret < 0)
+	continue;
+
+      break;
+    }
+
+  if (i == (sizeof(cover_extension) / sizeof(cover_extension[0])))
+    return -1;
+
+  DPRINTF(E_DBG, L_ART, "Found parent directory artwork file %s\n", artwork);
+
+  return artwork_get(artwork, max_w, max_h, format, evbuf);
+}
+
+
 
 int
 artwork_get_item_filename(char *filename, int max_w, int max_h, int format, struct evbuffer *evbuf)
 {
   int ret;
 
-  /* FUTURE: look at embedded artwork */
+#if LIBAVFORMAT_VERSION_MAJOR >= 55 || (LIBAVFORMAT_VERSION_MAJOR == 54 && LIBAVFORMAT_VERSION_MINOR >= 6)
+  /* Look for embedded artwork */
+  ret = artwork_get_embedded_image(filename, max_w, max_h, format, evbuf);
+  if (ret > 0)
+    return ret;
+#endif
 
   /* Look for basename(filename).{png,jpg} */
   ret = artwork_get_own_image(filename, max_w, max_h, format, evbuf);
@@ -741,6 +1084,11 @@ artwork_get_item_filename(char *filename, int max_w, int max_h, int format, stru
   if (ret > 0)
     return ret;
 
+  /* Look for parentdir(filename).{png,jpg} */
+  ret = artwork_get_parentdir_image(filename, 0, max_w, max_h, format, evbuf);
+  if (ret > 0)
+    return ret;
+
   return -1;
 }
 
@@ -773,6 +1121,9 @@ artwork_get_group(int id, int max_w, int max_h, int format, struct evbuffer *evb
   char *dir;
   int got_art;
   int ret;
+#if LIBAVFORMAT_VERSION_MAJOR >= 55 || (LIBAVFORMAT_VERSION_MAJOR == 54 && LIBAVFORMAT_VERSION_MINOR >= 6)
+  int artwork_t;
+#endif
 
   DPRINTF(E_DBG, L_ART, "Artwork request for group %d\n", id);
 
@@ -791,10 +1142,11 @@ artwork_get_group(int id, int max_w, int max_h, int format, struct evbuffer *evb
       goto files_art;
     }
 
-  got_art = -1;
-  while ((got_art < 0) && ((ret = db_query_fetch_string(&qp, &dir)) == 0) && (dir))
+  got_art = 0;
+  while (!got_art && ((ret = db_query_fetch_string(&qp, &dir)) == 0) && (dir))
     {
-      got_art = artwork_get_dir_image(dir, 1, max_w, max_h, format, evbuf);
+      got_art = (artwork_get_dir_image(dir, 1, max_w, max_h, format, evbuf) > 0)
+                || (artwork_get_parentdir_image(dir, 1, max_w, max_h, format, evbuf) > 0);
     }
 
   db_query_end(&qp);
@@ -809,29 +1161,38 @@ artwork_get_group(int id, int max_w, int max_h, int format, struct evbuffer *evb
  files_art:
   memset(&qp, 0, sizeof(struct query_params));
 
-  qp.type = Q_GROUPITEMS;
+  qp.type = Q_GROUP_ITEMS;
   qp.id = id;
 
   ret = db_query_start(&qp);
   if (ret < 0)
     {
-      DPRINTF(E_LOG, L_ART, "Could not start Q_GROUPITEMS query\n");
+      DPRINTF(E_LOG, L_ART, "Could not start Q_GROUP_ITEMS query\n");
 
       return -1;
     }
 
-  got_art = -1;
-  while ((got_art < 0) && ((ret = db_query_fetch_file(&qp, &dbmfi)) == 0) && (dbmfi.id))
+  got_art = 0;
+  while (!got_art && ((ret = db_query_fetch_file(&qp, &dbmfi)) == 0) && (dbmfi.id))
     {
-      got_art = artwork_get_own_image(dbmfi.path, max_w, max_h, format, evbuf);
+#if LIBAVFORMAT_VERSION_MAJOR >= 55 || (LIBAVFORMAT_VERSION_MAJOR == 54 && LIBAVFORMAT_VERSION_MINOR >= 6)
+      if ((safe_atoi32(dbmfi.artwork, &artwork_t) == 0) && (artwork_t == ARTWORK_EMBEDDED))
+	got_art = (artwork_get_embedded_image(dbmfi.path, max_w, max_h, format, evbuf) > 0);
+      else
+	got_art = (artwork_get_own_image(dbmfi.path, max_w, max_h, format, evbuf) > 0);
+#else
+      got_art = (artwork_get_own_image(dbmfi.path, max_w, max_h, format, evbuf) > 0);
+#endif
     }
 
   db_query_end(&qp);
 
   if (ret < 0)
-    DPRINTF(E_LOG, L_ART, "Error fetching Q_GROUPITEMS results\n");
+    DPRINTF(E_LOG, L_ART, "Error fetching Q_GROUP_ITEMS results\n");
   else if (got_art > 0)
     return got_art;
 
+  DPRINTF(E_DBG, L_ART, "No artwork found for group %d\n", id);
+
   return -1;
 }
