@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2009-2010 Julien BLACHE <jb@jblache.org>
+ * Copyright (C) 2009-2011 Julien BLACHE <jb@jblache.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,7 +25,9 @@
 #include <stdlib.h>
 #include <stdint.h>
 
+#include <sys/types.h>
 #include <sys/utsname.h>
+#include <pwd.h>
 
 #include <errno.h>
 
@@ -47,7 +49,7 @@ static cfg_opt_t sec_general[] =
     CFG_STR("logfile", STATEDIR "/log/" PACKAGE ".log", CFGF_NONE),
     CFG_STR("db_path", STATEDIR "/cache/" PACKAGE "/songs3.db", CFGF_NONE),
     CFG_INT_CB("loglevel", E_LOG, CFGF_NONE, &cb_loglevel),
-    CFG_BOOL("ipv6", cfg_true, CFGF_NONE),
+    CFG_BOOL("ipv6", cfg_false, CFGF_NONE),
     CFG_END()
   };
 
@@ -58,7 +60,19 @@ static cfg_opt_t sec_library[] =
     CFG_INT("port", 3689, CFGF_NONE),
     CFG_STR("password", NULL, CFGF_NONE),
     CFG_STR_LIST("directories", NULL, CFGF_NONE),
+    CFG_STR_LIST("podcasts", NULL, CFGF_NONE),
+    CFG_STR_LIST("audiobooks", NULL, CFGF_NONE),
     CFG_STR_LIST("compilations", NULL, CFGF_NONE),
+    CFG_STR("compilation_artist", NULL, CFGF_NONE),
+    CFG_STR("name_library", "Library", CFGF_NONE),
+    CFG_STR("name_music", "Music", CFGF_NONE),
+    CFG_STR("name_movies", "Movies", CFGF_NONE),
+    CFG_STR("name_tvshows", "TV Shows", CFGF_NONE),
+    CFG_STR("name_podcasts", "Podcasts", CFGF_NONE),
+    CFG_STR("name_audiobooks", "Audiobooks", CFGF_NONE),
+    CFG_STR_LIST("artwork_basenames", "{artwork,cover,Folder}", CFGF_NONE),
+    CFG_STR_LIST("filetypes_ignore", "{.db,.ini}", CFGF_NONE),
+    CFG_BOOL("filescan_disable", cfg_false, CFGF_NONE),
     CFG_BOOL("itunes_overrides", cfg_false, CFGF_NONE),
     CFG_STR_LIST("no_transcode", NULL, CFGF_NONE),
     CFG_STR_LIST("force_transcode", NULL, CFGF_NONE),
@@ -74,28 +88,42 @@ static cfg_opt_t sec_audio[] =
 #elif defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
     CFG_STR("card", "/dev/dsp", CFGF_NONE),
 #endif
+    CFG_STR("mixer", NULL, CFGF_NONE),
     CFG_END()
   };
 
-/* ApEx device section structure */
-static cfg_opt_t sec_apex[] =
+/* AirPlay/ApEx device section structure */
+static cfg_opt_t sec_airplay[] =
   {
+    CFG_INT("max_volume", 11, CFGF_NONE),
     CFG_STR("password", NULL, CFGF_NONE),
     CFG_END()
   };
 
+/* Spotify section structure */
+static cfg_opt_t sec_spotify[] =
+  {
+    CFG_STR("settings_dir", STATEDIR "/cache/" PACKAGE "/libspotify", CFGF_NONE),
+    CFG_STR("cache_dir", "/tmp", CFGF_NONE),
+    CFG_INT("bitrate", 0, CFGF_NONE),
+    CFG_END()
+  };
+
 /* Config file structure */
 static cfg_opt_t toplvl_cfg[] =
   {
     CFG_SEC("general", sec_general, CFGF_NONE),
     CFG_SEC("library", sec_library, CFGF_NONE),
     CFG_SEC("audio", sec_audio, CFGF_NONE),
-    CFG_SEC("apex", sec_apex, CFGF_MULTI | CFGF_TITLE),
+    CFG_SEC("airplay", sec_airplay, CFGF_MULTI | CFGF_TITLE),
+    CFG_SEC("spotify", sec_spotify, CFGF_NONE),
     CFG_END()
   };
 
 cfg_t *cfg;
 uint64_t libhash;
+uid_t runas_uid;
+gid_t runas_gid;
 
 
 static int
@@ -241,6 +269,8 @@ int
 conffile_load(char *file)
 {
   cfg_t *lib;
+  struct passwd *pw;
+  char *runas;
   int ret;
 
   cfg = cfg_init(toplvl_cfg, CFGF_NONE);
@@ -260,6 +290,19 @@ conffile_load(char *file)
       goto out_fail;
     }
 
+  /* Resolve runas username */
+  runas = cfg_getstr(cfg_getsec(cfg, "general"), "uid");
+  pw = getpwnam(runas);
+  if (!pw)
+    {
+      DPRINTF(E_FATAL, L_CONF, "Could not lookup user %s: %s\n", runas, strerror(errno));
+
+      goto out_fail;
+    }
+
+  runas_uid = pw->pw_uid;
+  runas_gid = pw->pw_gid;
+
   lib = cfg_getsec(cfg, "library");
 
   if (cfg_size(lib, "directories") == 0)
