@@ -23,9 +23,6 @@
 #include <string.h>
 #include <stdint.h>
 
-#include <event.h>
-#include "evhttp/evhttp.h"
-
 #include "db.h"
 #include "misc.h"
 #include "logger.h"
@@ -167,6 +164,20 @@ dmap_add_literal(struct evbuffer *evbuf, char *tag, char *str, int len)
     evbuffer_add(evbuf, str, len);
 }
 
+void
+dmap_add_raw_uint32(struct evbuffer *evbuf, uint32_t val)
+{
+  unsigned char buf[4];
+
+  /* Value */
+  buf[0] = (val >> 24) & 0xff;
+  buf[1] = (val >> 16) & 0xff;
+  buf[2] = (val >> 8) & 0xff;
+  buf[3] = val & 0xff;
+
+  evbuffer_add(evbuf, buf, sizeof(buf));
+}
+
 void
 dmap_add_string(struct evbuffer *evbuf, char *tag, const char *str)
 {
@@ -378,11 +389,13 @@ dmap_encode_file_metadata(struct evbuffer *songlist, struct evbuffer *song, stru
   int32_t val;
   int want_mikd;
   int want_asdk;
+  int want_ased;
   int i;
   int ret;
 
   want_mikd = 0;
   want_asdk = 0;
+  want_ased = 0;
 
   i = -1;
   while (1)
@@ -396,7 +409,10 @@ dmap_encode_file_metadata(struct evbuffer *songlist, struct evbuffer *song, stru
 	    break;
 
 	  df = meta[i];
-	  dfm = df->dfm;
+	  if (df->dfm)
+	    dfm = df->dfm;
+	  else
+	    break;
 	}
       /* No specific meta tags requested, send out everything */
       else
@@ -409,6 +425,13 @@ dmap_encode_file_metadata(struct evbuffer *songlist, struct evbuffer *song, stru
 	  dfm = dmap_fields[i].dfm;
 	}
 
+      /* Extradata not in media_file_info but flag for reply */
+      if (dfm == &dfm_dmap_ased)
+	{
+	  want_ased = 1;
+	  continue;
+	}
+
       /* Not in struct media_file_info */
       if (dfm->mfi_offset < 0)
 	continue;
@@ -427,7 +450,7 @@ dmap_encode_file_metadata(struct evbuffer *songlist, struct evbuffer *song, stru
 	  continue;
 	}
 
-      DPRINTF(E_DBG, L_DAAP, "Investigating %s\n", df->desc);
+      DPRINTF(E_SPAM, L_DAAP, "Investigating %s\n", df->desc);
 
       strval = (char **) ((char *)dbmfi + dfm->mfi_offset);
 
@@ -476,7 +499,14 @@ dmap_encode_file_metadata(struct evbuffer *songlist, struct evbuffer *song, stru
 
       dmap_add_field(song, df, *strval, val);
 
-      DPRINTF(E_DBG, L_DAAP, "Done with meta tag %s (%s)\n", df->desc, *strval);
+      DPRINTF(E_SPAM, L_DAAP, "Done with meta tag %s (%s)\n", df->desc, *strval);
+    }
+
+  /* Required for artwork in iTunes, set songartworkcount (asac) = 1 */
+  if (want_ased)
+    {
+      dmap_add_short(song, "ased", 1);
+      dmap_add_short(song, "asac", 1);
     }
 
   if (sort_tags)
