@@ -1656,6 +1656,14 @@ evhttp_read_header(struct evhttp_connection *evcon, struct evhttp_request *req)
 		return;
 	}
 
+	/* Callback can shut down connection with negative return value */
+	if (req->header_cb != NULL) {
+		if ((*req->header_cb)(req, req->cb_arg) < 0) {
+			evhttp_connection_fail(evcon, EVCON_HTTP_EOF);
+			return;
+		}
+	}
+
 	/* Done reading headers, do the real work */
 	switch (req->kind) {
 	case EVHTTP_REQUEST:
@@ -2006,7 +2014,9 @@ evhttp_send_reply_start(struct evhttp_request *req, int code,
 	evhttp_connection_stop_detectclose(req->evcon);
 
 	evhttp_response_code(req, code, reason);
-	if (req->major == 1 && req->minor == 1) {
+	/* If the client is requesting a partial file, don't use
+	 * chunked encoding, or video files won't work in iTunes. */
+	if (req->major == 1 && req->minor == 1 && code == 200) {
 		/* use chunked encoding for HTTP/1.1 */
 		evhttp_add_header(req->output_headers, "Transfer-Encoding",
 		    "chunked");
