@@ -37,6 +37,8 @@
 #include <dirent.h>
 #include <pthread.h>
 
+#include <unistr.h>
+#include <unictype.h>
 #include <uninorm.h>
 
 #if defined(__linux__)
@@ -59,13 +61,21 @@
 #include "conffile.h"
 #include "misc.h"
 #include "remote_pairing.h"
+#include "player.h"
+
+#ifdef HAVE_SPOTIFY_H
+# include "spotify.h"
+#endif
 
 
 #define F_SCAN_BULK    (1 << 0)
 #define F_SCAN_RESCAN  (1 << 1)
+#define F_SCAN_FAST    (1 << 2)
+#define F_SCAN_MOVED   (1 << 3)
 
 struct deferred_pl {
   char *path;
+  time_t mtime;
   struct deferred_pl *next;
 };
 
@@ -89,6 +99,16 @@ static pthread_t tid_scan;
 static struct deferred_pl *playlists;
 static struct stacked_dir *dirstack;
 
+/* Count of files scanned during a bulk scan */
+static int counter;
+
+/* Forward */
+static void
+bulk_scan(int flags);
+static int
+inofd_event_set(void);
+static void
+inofd_event_unset(void);
 
 static int
 push_dir(struct stacked_dir **s, char *path)
@@ -133,31 +153,129 @@ pop_dir(struct stacked_dir **s)
   return ret;
 }
 
+static int
+ignore_filetype(char *ext)
+{
+  cfg_t *lib;
+  int n;
+  int i;
+
+  lib = cfg_getsec(cfg, "library");
+  n = cfg_size(lib, "filetypes_ignore");
+
+  for (i = 0; i < n; i++)
+    {
+      if (strcasecmp(ext, cfg_getnstr(lib, "filetypes_ignore", i)) == 0)
+	return 1;
+    }
+
+  return 0;
+}
 
 static void
-normalize_fixup_tag(char **tag, char *src_tag)
+sort_tag_create(char **sort_tag, char *src_tag)
 {
-  char *norm;
+  const uint8_t *i_ptr;
+  const uint8_t *n_ptr;
+  const uint8_t *number;
+  uint8_t out[1024];
+  uint8_t *o_ptr;
+  int append_number;
+  ucs4_t puc;
+  int numlen;
   size_t len;
 
   /* Note: include terminating NUL in string length for u8_normalize */
 
-  if (!*tag)
-    *tag = (char *)u8_normalize(UNINORM_NFD, (uint8_t *)src_tag, strlen(src_tag) + 1, NULL, &len);
-  else
+  if (*sort_tag)
     {
-      norm = (char *)u8_normalize(UNINORM_NFD, (uint8_t *)*tag, strlen(*tag) + 1, NULL, &len);
-      free(*tag);
-      *tag = norm;
+      DPRINTF(E_DBG, L_SCAN, "Existing sort tag will be normalized: %s\n", *sort_tag);
+      o_ptr = u8_normalize(UNINORM_NFD, (uint8_t *)*sort_tag, strlen(*sort_tag) + 1, NULL, &len);
+      free(*sort_tag);
+      *sort_tag = (char *)o_ptr;
+      return;
     }
+
+  if (!src_tag || ((len = strlen(src_tag)) == 0))
+    {
+      *sort_tag = NULL;
+      return;
+    }
+
+  // Set input pointer past article if present
+  if ((strncasecmp(src_tag, "a ", 2) == 0) && (len > 2))
+    i_ptr = (uint8_t *)(src_tag + 2);
+  else if ((strncasecmp(src_tag, "an ", 3) == 0) && (len > 3))
+    i_ptr = (uint8_t *)(src_tag + 3);
+  else if ((strncasecmp(src_tag, "the ", 4) == 0) && (len > 4))
+    i_ptr = (uint8_t *)(src_tag + 4);
+  else
+    i_ptr = (uint8_t *)src_tag;
+
+  // Poor man's natural sort. Makes sure we sort like this: a1, a2, a10, a11, a21, a111
+  // We do this by padding zeroes to (short) numbers. As an alternative we could have
+  // made a proper natural sort algorithm in sqlext.c, but we don't, since we don't
+  // want any risk of hurting response times
+  memset(&out, 0, sizeof(out));
+  o_ptr = (uint8_t *)&out;
+  number = NULL;
+  append_number = 0;
+
+  do
+    {
+      n_ptr = u8_next(&puc, i_ptr);
+
+      if (uc_is_digit(puc))
+	{
+	  if (!number) // We have encountered the beginning of a number
+	    number = i_ptr;
+	  append_number = (n_ptr == NULL); // If last char in string append number now
+	}
+      else
+	{
+	  if (number)
+	    append_number = 1; // A number has ended so time to append it
+	  else
+	    o_ptr = u8_stpncpy(o_ptr, i_ptr, u8_strmblen(i_ptr)); // No numbers in sight, just append char
+	}
+
+      // Break if less than 100 bytes remain (prevent buffer overflow)
+      if (sizeof(out) - u8_strlen(out) < 100)
+	break;
+
+      // Break if number is very large (prevent buffer overflow)
+      if (number && (i_ptr - number > 50))
+	break;
+
+      if (append_number)
+	{
+	  numlen = i_ptr - number;
+	  if (numlen < 5) // Max pad width
+	    {
+	      u8_strcpy(o_ptr, (uint8_t *)"00000");
+	      o_ptr += (5 - numlen);
+	    }
+	  o_ptr = u8_stpncpy(o_ptr, number, numlen + u8_strmblen(i_ptr));
+
+	  number = NULL;
+	  append_number = 0;
+	}
+
+      i_ptr = n_ptr;
+    }
+  while (n_ptr);
+
+  *sort_tag = (char *)u8_normalize(UNINORM_NFD, (uint8_t *)&out, u8_strlen(out) + 1, NULL, &len);
 }
 
 static void
 fixup_tags(struct media_file_info *mfi)
 {
+  cfg_t *lib;
   size_t len;
   char *tag;
   char *sep = " - ";
+  char *ca;
 
   if (mfi->genre && (strlen(mfi->genre) == 0))
     {
@@ -181,7 +299,7 @@ fixup_tags(struct media_file_info *mfi)
    * Default to mpeg4 video/audio for unknown file types
    * in an attempt to allow streaming of DRM-afflicted files
    */
-  if (strcmp(mfi->codectype, "unkn") == 0)
+  if (mfi->codectype && strcmp(mfi->codectype, "unkn") == 0)
     {
       if (mfi->has_video)
 	{
@@ -266,154 +384,216 @@ fixup_tags(struct media_file_info *mfi)
 	mfi->title = strdup(mfi->fname);
     }
 
-  /* Ensure sort tags are filled and normalized */
-  normalize_fixup_tag(&mfi->artist_sort, mfi->artist);
-  normalize_fixup_tag(&mfi->album_sort, mfi->album);
-  normalize_fixup_tag(&mfi->title_sort, mfi->title);
+  /* Ensure sort tags are filled, manipulated and normalized */
+  sort_tag_create(&mfi->artist_sort, mfi->artist);
+  sort_tag_create(&mfi->album_sort, mfi->album);
+  sort_tag_create(&mfi->title_sort, mfi->title);
 
-  /* If we don't have an album_artist, set it to artist */
-  if (!mfi->album_artist)
+  /* We need to set album_artist according to media type and config */
+  if (mfi->compilation)          /* Compilation */
     {
-      if (mfi->compilation)
+      lib = cfg_getsec(cfg, "library");
+      ca = cfg_getstr(lib, "compilation_artist");
+      if (ca && mfi->album_artist)
+	{
+	  free(mfi->album_artist);
+	  mfi->album_artist = strdup(ca);
+	}
+      else if (ca && !mfi->album_artist)
+	{
+	  mfi->album_artist = strdup(ca);
+	}
+      else if (!ca && !mfi->album_artist)
 	{
 	  mfi->album_artist = strdup("");
 	  mfi->album_artist_sort = strdup("");
 	}
-      else
-	mfi->album_artist = strdup(mfi->artist);
+    }
+  else if (mfi->media_kind == 4) /* Podcast */
+    {
+      if (mfi->album_artist)
+	free(mfi->album_artist);
+      mfi->album_artist = strdup("");
+      mfi->album_artist_sort = strdup("");
+    }
+  else if (!mfi->album_artist)   /* Regular media without album_artist */
+    {
+      mfi->album_artist = strdup(mfi->artist);
     }
 
   if (!mfi->album_artist_sort && (strcmp(mfi->album_artist, mfi->artist) == 0))
     mfi->album_artist_sort = strdup(mfi->artist_sort);
   else
-    normalize_fixup_tag(&mfi->album_artist_sort, mfi->album_artist);
+    sort_tag_create(&mfi->album_artist_sort, mfi->album_artist);
 
   /* Composer is not one of our mandatory tags, so take extra care */
   if (mfi->composer_sort || mfi->composer)
-    normalize_fixup_tag(&mfi->composer_sort, mfi->composer);
+    sort_tag_create(&mfi->composer_sort, mfi->composer);
 }
 
 
-static void
-process_media_file(char *file, time_t mtime, off_t size, int compilation)
+void
+filescanner_process_media(char *path, time_t mtime, off_t size, int type, struct media_file_info *external_mfi)
 {
-  struct media_file_info mfi;
+  struct media_file_info *mfi;
   char *filename;
   char *ext;
   time_t stamp;
   int id;
   int ret;
 
-  db_file_stamp_bypath(file, &stamp, &id);
+  filename = strrchr(path, '/');
+  if ((!filename) || (strlen(filename) == 1))
+    filename = path;
+  else
+    filename++;
 
-  if (stamp >= mtime)
+  /* File types which should never be processed */
+  ext = strrchr(path, '.');
+  if (ext)
+    {
+      if ((strcasecmp(ext, ".pls") == 0) || (strcasecmp(ext, ".url") == 0))
+	{
+	  DPRINTF(E_INFO, L_SCAN, "No support for .url and .pls in this version, use .m3u\n");
+
+	  return;
+	}
+      else if ((strcasecmp(ext, ".png") == 0) || (strcasecmp(ext, ".jpg") == 0))
+	{
+	  /* Artwork files - don't scan */
+	  return;
+	}
+      else if ((filename[0] == '_') || (filename[0] == '.'))
+	{
+	  /* Hidden files - don't scan */
+	  return;
+	}
+      else if (ignore_filetype(ext))
+	{
+	  /* File extension is in ignore list - don't scan */
+	  return;
+	}
+    }
+
+  db_file_stamp_bypath(path, &stamp, &id);
+
+  if (stamp && (stamp >= mtime))
     {
       db_file_ping(id);
       return;
     }
 
-  memset(&mfi, 0, sizeof(struct media_file_info));
+  if (!external_mfi)
+    {
+      mfi = (struct media_file_info*)malloc(sizeof(struct media_file_info));
+      if (!mfi)
+	{
+	  DPRINTF(E_LOG, L_SCAN, "Out of memory for mfi\n");
+	  return;
+	}
+
+      memset(mfi, 0, sizeof(struct media_file_info));
+    }
+  else
+    mfi = external_mfi;
 
   if (stamp)
-    mfi.id = db_file_id_bypath(file);
+    mfi->id = db_file_id_bypath(path);
 
-  filename = strrchr(file, '/');
-  if (!filename)
+  mfi->fname = strdup(filename);
+  if (!mfi->fname)
     {
-      DPRINTF(E_LOG, L_SCAN, "Could not determine filename for %s\n", file);
-
-      return;
-    }
-
-  mfi.fname = strdup(filename + 1);
-  if (!mfi.fname)
-    {
-      DPRINTF(E_WARN, L_SCAN, "Out of memory for fname\n");
-
-      return;
-    }
-
-  mfi.path = strdup(file);
-  if (!mfi.path)
-    {
-      DPRINTF(E_WARN, L_SCAN, "Out of memory for path\n");
-
-      free(mfi.fname);
-      return;
-    }
-
-  mfi.time_modified = mtime;
-  mfi.file_size = size;
-
-  ret = -1;
-
-  /* Special cases */
-  ext = strrchr(file, '.');
-  if (ext)
-    {
-      if ((strcmp(ext, ".pls") == 0)
-	  || (strcmp(ext, ".url") == 0))
-	{
-	  mfi.data_kind = 1; /* url/stream */
-
-	  ret = scan_url_file(file, &mfi);
-	  if (ret < 0)
-	    goto out;
-	}
-      else if ((strcmp(ext, ".png") == 0)
-	       || (strcmp(ext, ".jpg") == 0))
-	{
-	  /* Artwork - don't scan */
-	  goto out;
-	}
-    }
-
-  /* General case */
-  if (ret < 0)
-    {
-      ret = scan_metadata_ffmpeg(file, &mfi);
-      mfi.data_kind = 0; /* real file */
-    }
-
-  if (ret < 0)
-    {
-      DPRINTF(E_INFO, L_SCAN, "Could not extract metadata for %s\n", file);
-
+      DPRINTF(E_LOG, L_SCAN, "Out of memory for fname\n");
       goto out;
     }
 
-  mfi.compilation = compilation;
+  mfi->path = strdup(path);
+  if (!mfi->path)
+    {
+      DPRINTF(E_LOG, L_SCAN, "Out of memory for path\n");
+      goto out;
+    }
 
-  if (!mfi.item_kind)
-    mfi.item_kind = 2; /* music */
-  if (!mfi.media_kind)
-    mfi.media_kind = 1; /* music */
+  mfi->time_modified = mtime;
+  mfi->file_size = size;
 
-  unicode_fixup_mfi(&mfi);
-
-  fixup_tags(&mfi);
-
-  if (mfi.id == 0)
-    db_file_add(&mfi);
+  if (type & F_SCAN_TYPE_FILE)
+    {
+      mfi->data_kind = 0; /* real file */
+      ret = scan_metadata_ffmpeg(path, mfi);
+    }
+  else if (type & F_SCAN_TYPE_URL)
+    {
+      mfi->data_kind = 1; /* url/stream */
+#if LIBAVFORMAT_VERSION_MAJOR >= 56 || (LIBAVFORMAT_VERSION_MAJOR == 55 && LIBAVFORMAT_VERSION_MINOR >= 13)
+      ret = scan_metadata_ffmpeg(path, mfi);
+#else
+      ret = scan_metadata_icy(path, mfi);
+#endif
+    }
+  else if (type & F_SCAN_TYPE_SPOTIFY)
+    {
+      mfi->data_kind = 2; /* iTunes has no spotify data kind, but we use 2 */
+      ret = mfi->artist && mfi->album && mfi->title;
+    }
+  else if (type & F_SCAN_TYPE_PIPE)
+    {
+      mfi->data_kind = 3; /* iTunes has no pipe data kind, but we use 3 */
+      mfi->type = strdup("wav");
+      mfi->codectype = strdup("wav");
+      mfi->description = strdup("PCM16 pipe");
+      ret = 1;
+    }
   else
-    db_file_update(&mfi);
+    {
+      DPRINTF(E_LOG, L_SCAN, "Unknown scan type for %s, this error should not occur\n", path);
+      ret = -1;
+    }
+
+  if (ret < 0)
+    {
+      DPRINTF(E_INFO, L_SCAN, "Could not extract metadata for %s\n", path);
+      goto out;
+    }
+
+  if (type & F_SCAN_TYPE_COMPILATION)
+    mfi->compilation = 1;
+  if (type & F_SCAN_TYPE_PODCAST)
+    mfi->media_kind = 4; /* podcast */
+  if (type & F_SCAN_TYPE_AUDIOBOOK)
+    mfi->media_kind = 8; /* audiobook */
+
+  if (!mfi->item_kind)
+    mfi->item_kind = 2; /* music */
+  if (!mfi->media_kind)
+    mfi->media_kind = 1; /* music */
+
+  unicode_fixup_mfi(mfi);
+
+  fixup_tags(mfi);
+
+  if (mfi->id == 0)
+    db_file_add(mfi);
+  else
+    db_file_update(mfi);
 
  out:
-  free_mfi(&mfi, 1);
+  if (!external_mfi)
+    free_mfi(mfi, 0);
 }
 
 static void
-process_playlist(char *file)
+process_playlist(char *file, time_t mtime)
 {
   char *ext;
 
   ext = strrchr(file, '.');
   if (ext)
     {
-      if (strcmp(ext, ".m3u") == 0)
-	scan_m3u_playlist(file);
+      if (strcasecmp(ext, ".m3u") == 0)
+	scan_m3u_playlist(file, mtime);
 #ifdef ITUNES
-      else if (strcmp(ext, ".xml") == 0)
+      else if (strcasecmp(ext, ".xml") == 0)
 	scan_itunes_itml(file);
 #endif
     }
@@ -421,7 +601,7 @@ process_playlist(char *file)
 
 /* Thread: scan */
 static void
-defer_playlist(char *path)
+defer_playlist(char *path, time_t mtime)
 {
   struct deferred_pl *pl;
 
@@ -444,6 +624,7 @@ defer_playlist(char *path)
       return;
     }
 
+  pl->mtime = mtime;
   pl->next = playlists;
   playlists = pl;
 
@@ -460,7 +641,7 @@ process_deferred_playlists(void)
     {
       playlists = pl->next;
 
-      process_playlist(pl->path);
+      process_playlist(pl->path, pl->mtime);
 
       free(pl->path);
       free(pl);
@@ -475,23 +656,23 @@ process_deferred_playlists(void)
 
 /* Thread: scan */
 static void
-process_file(char *file, time_t mtime, off_t size, int compilation, int flags)
+process_file(char *file, time_t mtime, off_t size, int type, int flags)
 {
   char *ext;
 
   ext = strrchr(file, '.');
   if (ext)
     {
-      if ((strcmp(ext, ".m3u") == 0)
+      if ((strcasecmp(ext, ".m3u") == 0)
 #ifdef ITUNES
-	  || (strcmp(ext, ".xml") == 0)
+	  || (strcasecmp(ext, ".xml") == 0)
 #endif
 	  )
 	{
 	  if (flags & F_SCAN_BULK)
-	    defer_playlist(file);
+	    defer_playlist(file, mtime);
 	  else
-	    process_playlist(file);
+	    process_playlist(file, mtime);
 
 	  return;
 	}
@@ -501,27 +682,77 @@ process_file(char *file, time_t mtime, off_t size, int compilation, int flags)
 
 	  return;
 	}
+#ifdef HAVE_SPOTIFY_H
+      else if (strcmp(ext, ".spotify") == 0)
+	{
+	  spotify_login(file);
+
+	  return;
+	}
+#endif
+      else if (strcmp(ext, ".full-rescan") == 0)
+	{
+	  if (flags & F_SCAN_BULK)
+	    return;
+	  else
+	    {
+	      DPRINTF(E_LOG, L_SCAN, "Forcing full rescan, found full-rescan file: %s\n", file);
+	      player_playback_stop();
+	      player_queue_clear();
+	      inofd_event_unset(); // Clears all inotify watches
+	      db_purge_all(); // Clears files, playlists, playlistitems, inotify and groups
+
+	      inofd_event_set();
+	      bulk_scan(F_SCAN_BULK);
+
+	      return;
+	    }
+	}
+      else if (strcmp(ext, ".init-rescan") == 0)
+	{
+	  if (flags & F_SCAN_BULK)
+	    return;
+	  else
+	    {
+	      DPRINTF(E_LOG, L_SCAN, "Forcing startup rescan, found init-rescan file: %s\n", file);
+	      inofd_event_unset(); // Clears all inotify watches
+	      db_watch_clear();
+
+	      inofd_event_set();
+	      bulk_scan(F_SCAN_BULK | F_SCAN_RESCAN);
+
+	      return;
+	    }
+	}
     }
 
   /* Not any kind of special file, so let's see if it's a media file */
-  process_media_file(file, mtime, size, compilation);
-}
+  filescanner_process_media(file, mtime, size, type, NULL);
 
+  counter++;
+
+  /* When in bulk mode, split transaction in pieces of 200 */
+  if ((flags & F_SCAN_BULK) && (counter % 200 == 0))
+    {
+      DPRINTF(E_LOG, L_SCAN, "Scanned %d files...\n", counter);
+      db_transaction_end();
+      db_transaction_begin();
+    }
+}
 
 /* Thread: scan */
 static int
-check_compilation(char *path)
+check_speciallib(char *path, const char *libtype)
 {
   cfg_t *lib;
   int ndirs;
   int i;
 
   lib = cfg_getsec(cfg, "library");
-  ndirs = cfg_size(lib, "compilations");
-
+  ndirs = cfg_size(lib, libtype);
   for (i = 0; i < ndirs; i++)
     {
-      if (strstr(path, cfg_getnstr(lib, "compilations", i)))
+      if (strstr(path, cfg_getnstr(lib, libtype, i)))
 	return 1;
     }
 
@@ -543,7 +774,7 @@ process_directory(char *path, int flags)
 #if defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
   struct kevent kev;
 #endif
-  int compilation;
+  int type;
   int ret;
 
   if (flags & F_SCAN_BULK)
@@ -574,8 +805,14 @@ process_directory(char *path, int flags)
       return;
     }
 
-  /* Check for a compilation directory */
-  compilation = check_compilation(path);
+  /* Check if compilation and/or podcast directory */
+  type = 0;
+  if (check_speciallib(path, "compilations"))
+    type |= F_SCAN_TYPE_COMPILATION;
+  if (check_speciallib(path, "podcasts"))
+    type |= F_SCAN_TYPE_PODCAST;
+  if (check_speciallib(path, "audiobooks"))
+    type |= F_SCAN_TYPE_AUDIOBOOK;
 
   for (;;)
     {
@@ -639,11 +876,19 @@ process_directory(char *path, int flags)
 	}
 
       if (S_ISREG(sb.st_mode))
-	process_file(entry, sb.st_mtime, sb.st_size, compilation, flags);
+	{
+	  if (!(flags & F_SCAN_FAST))
+	    process_file(entry, sb.st_mtime, sb.st_size, F_SCAN_TYPE_FILE | type, flags);
+	}
+      else if (S_ISFIFO(sb.st_mode))
+	{
+	  if (!(flags & F_SCAN_FAST))
+	    process_file(entry, sb.st_mtime, sb.st_size, F_SCAN_TYPE_PIPE | type, flags);
+	}
       else if (S_ISDIR(sb.st_mode))
 	push_dir(&dirstack, entry);
       else
-	DPRINTF(E_LOG, L_SCAN, "Skipping %s, not a directory, symlink nor regular file\n", entry);
+	DPRINTF(E_LOG, L_SCAN, "Skipping %s, not a directory, symlink, pipe nor regular file\n", entry);
     }
 
   closedir(dirp);
@@ -652,7 +897,7 @@ process_directory(char *path, int flags)
 
 #if defined(__linux__)
   /* Add inotify watch */
-  wi.wd = inotify_add_watch(inofd, path, IN_CREATE | IN_DELETE | IN_MODIFY | IN_CLOSE_WRITE | IN_MOVE | IN_DELETE | IN_MOVE_SELF);
+  wi.wd = inotify_add_watch(inofd, path, IN_ATTRIB | IN_CREATE | IN_DELETE | IN_CLOSE_WRITE | IN_MOVE | IN_DELETE | IN_MOVE_SELF);
   if (wi.wd < 0)
     {
       DPRINTF(E_WARN, L_SCAN, "Could not create inotify watch for %s: %s\n", path, strerror(errno));
@@ -660,7 +905,7 @@ process_directory(char *path, int flags)
       return;
     }
 
-  if (!(flags & F_SCAN_RESCAN))
+  if (!(flags & F_SCAN_MOVED))
     {
       wi.cookie = 0;
       wi.path = path;
@@ -723,13 +968,14 @@ process_directories(char *root, int flags)
 
 /* Thread: scan */
 static void
-bulk_scan(void)
+bulk_scan(int flags)
 {
   cfg_t *lib;
   int ndirs;
   char *path;
   char *deref;
   time_t start;
+  time_t end;
   int i;
 
   start = time(NULL);
@@ -749,10 +995,20 @@ bulk_scan(void)
 	{
 	  DPRINTF(E_LOG, L_SCAN, "Skipping library directory %s, could not dereference: %s\n", path, strerror(errno));
 
+	  /* Assume dir is mistakenly not mounted, so just disable everything and update timestamps */
+	  db_file_disable_bymatch(path, "", 0);
+	  db_pl_disable_bymatch(path, "", 0);
+
+	  db_file_ping_bymatch(path, 1);
+	  db_pl_ping_bymatch(path, 1);
+
 	  continue;
 	}
 
-      process_directories(deref, F_SCAN_BULK);
+      counter = 0;
+      db_transaction_begin();
+      process_directories(deref, flags);
+      db_transaction_end();
 
       free(deref);
 
@@ -760,7 +1016,7 @@ bulk_scan(void)
 	return;
     }
 
-  if (playlists)
+  if (!(flags & F_SCAN_FAST) && playlists)
     process_deferred_playlists();
 
   if (scan_exit)
@@ -769,8 +1025,29 @@ bulk_scan(void)
   if (dirstack)
     DPRINTF(E_LOG, L_SCAN, "WARNING: unhandled leftover directories\n");
 
-  DPRINTF(E_DBG, L_SCAN, "Purging old database content\n");
-  db_purge_cruft(start);
+  end = time(NULL);
+
+  if (flags & F_SCAN_FAST)
+    {
+      DPRINTF(E_LOG, L_SCAN, "Bulk library scan completed in %.f sec (with file scan disabled)\n", difftime(end, start));
+    }
+  else
+    {
+      /* Protect spotify from the imminent purge if rescanning */
+      if (flags & F_SCAN_RESCAN)
+	{
+	  db_file_ping_bymatch("spotify:", 0);
+	  db_pl_ping_bymatch("spotify:", 0);
+	}
+
+      DPRINTF(E_DBG, L_SCAN, "Purging old database content\n");
+      db_purge_cruft(start);
+
+      DPRINTF(E_LOG, L_SCAN, "Bulk library scan completed in %.f sec\n", difftime(end, start));
+
+      DPRINTF(E_DBG, L_SCAN, "Running post library scan jobs\n");
+      db_hook_post_scan();
+    }
 }
 
 
@@ -804,15 +1081,21 @@ filescanner(void *arg)
       pthread_exit(NULL);
     }
 
-  /* Recompute all songalbumids, in case the SQLite DB got transferred
+  /* Recompute all songartistids and songalbumids, in case the SQLite DB got transferred
    * to a different host; the hash is not portable.
    * It will also rebuild the groups we just cleared.
    */
+  db_files_update_songartistid();
   db_files_update_songalbumid();
 
-  bulk_scan();
+  if (cfg_getbool(cfg_getsec(cfg, "library"), "filescan_disable"))
+    bulk_scan(F_SCAN_BULK | F_SCAN_FAST);
+  else
+    bulk_scan(F_SCAN_BULK);
 
-  db_hook_post_scan();
+#ifdef HAVE_SPOTIFY_H
+  spotify_login(NULL);
+#endif
 
   if (!scan_exit)
     {
@@ -832,16 +1115,47 @@ filescanner(void *arg)
 
 
 #if defined(__linux__)
+static int
+watches_clear(uint32_t wd, char *path)
+{
+  struct watch_enum we;
+  uint32_t rm_wd;
+  int ret;
+
+  inotify_rm_watch(inofd, wd);
+  db_watch_delete_bywd(wd);
+
+  memset(&we, 0, sizeof(struct watch_enum));
+
+  we.match = path;
+
+  ret = db_watch_enum_start(&we);
+  if (ret < 0)
+    return -1;
+
+  while ((db_watch_enum_fetchwd(&we, &rm_wd) == 0) && (rm_wd))
+    {
+      inotify_rm_watch(inofd, rm_wd);
+    }
+
+  db_watch_enum_end(&we);
+
+  db_watch_delete_bymatch(path);
+
+  return 0;
+}
+
 /* Thread: scan */
 static void
 process_inotify_dir(struct watch_info *wi, char *path, struct inotify_event *ie)
 {
   struct watch_enum we;
   uint32_t rm_wd;
+  char *s;
   int flags = 0;
   int ret;
 
-  DPRINTF(E_DBG, L_SCAN, "Directory event: 0x%x, cookie 0x%x, wd %d\n", ie->mask, ie->cookie, wi->wd);
+  DPRINTF(E_SPAM, L_SCAN, "Directory event: 0x%x, cookie 0x%x, wd %d\n", ie->mask, ie->cookie, wi->wd);
 
   if (ie->mask & IN_UNMOUNT)
     {
@@ -886,26 +1200,10 @@ process_inotify_dir(struct watch_info *wi, char *path, struct inotify_event *ie)
 	   * and we can't tell where it's going
 	   */
 
-	  inotify_rm_watch(inofd, ie->wd);
-	  db_watch_delete_bywd(ie->wd);
-
-	  memset(&we, 0, sizeof(struct watch_enum));
-
-	  we.match = path;
-
-	  ret = db_watch_enum_start(&we);
+	  ret = watches_clear(ie->wd, path);
 	  if (ret < 0)
 	    return;
 
-	  while ((db_watch_enum_fetchwd(&we, &rm_wd) == 0) && (rm_wd))
-	    {
-	      inotify_rm_watch(inofd, rm_wd);
-	    }
-
-	  db_watch_enum_end(&we);
-
-	  db_watch_delete_bymatch(path);
-
 	  db_file_disable_bymatch(path, "", 0);
 	  db_pl_disable_bymatch(path, "", 0);
 	}
@@ -928,12 +1226,46 @@ process_inotify_dir(struct watch_info *wi, char *path, struct inotify_event *ie)
 	  db_pl_enable_bycookie(ie->cookie, path);
 
 	  /* We'll rescan the directory tree to update playlists */
-	  flags |= F_SCAN_RESCAN;
+	  flags |= F_SCAN_MOVED;
 	}
 
       ie->mask |= IN_CREATE;
     }
 
+  if (ie->mask & IN_ATTRIB)
+    {
+      DPRINTF(E_DBG, L_SCAN, "Directory permissions changed (%s): %s\n", wi->path, path);
+
+      // Find out if we are already watching the dir (ret will be 0)
+      s = wi->path;
+      wi->path = path;
+      ret = db_watch_get_bypath(wi);
+      if (ret == 0)
+	free(wi->path);
+      wi->path = s;
+
+      if (euidaccess(path, (R_OK | X_OK)) < 0)
+	{
+	  DPRINTF(E_LOG, L_SCAN, "Directory access to '%s' failed: %s\n", path, strerror(errno));
+
+	  if (ret == 0)
+	    watches_clear(wi->wd, path);
+
+	  db_file_disable_bymatch(path, "", 0);
+	  db_pl_disable_bymatch(path, "", 0);
+	}
+      else if (ret < 0)
+	{
+	  DPRINTF(E_LOG, L_SCAN, "Directory access to '%s' achieved\n", path);
+
+	  ie->mask |= IN_CREATE;
+	}
+      else
+	{
+	  DPRINTF(E_INFO, L_SCAN, "Directory event, but '%s' already being watched\n", path);
+	}
+    }
+
   if (ie->mask & IN_CREATE)
     {
       process_directories(path, flags);
@@ -950,26 +1282,50 @@ process_inotify_file(struct watch_info *wi, char *path, struct inotify_event *ie
   struct stat sb;
   char *deref = NULL;
   char *file = path;
-  int compilation;
+  int type;
   int ret;
 
-  DPRINTF(E_DBG, L_SCAN, "File event: 0x%x, cookie 0x%x, wd %d\n", ie->mask, ie->cookie, wi->wd);
+  DPRINTF(E_SPAM, L_SCAN, "File event: 0x%x, cookie 0x%x, wd %d\n", ie->mask, ie->cookie, wi->wd);
 
   if (ie->mask & IN_DELETE)
     {
+      DPRINTF(E_DBG, L_SCAN, "File deleted: %s\n", path);
+
       db_file_delete_bypath(path);
       db_pl_delete_bypath(path);
     }
 
   if (ie->mask & IN_MOVED_FROM)
     {
-      db_file_disable_bypath(path, wi->path, ie->cookie);
-      db_pl_disable_bypath(path, wi->path, ie->cookie);
+      DPRINTF(E_DBG, L_SCAN, "File moved from: %s\n", path);
+
+      db_file_disable_bypath(path, path, ie->cookie);
+      db_pl_disable_bypath(path, path, ie->cookie);
+    }
+
+  if (ie->mask & IN_ATTRIB)
+    {
+      DPRINTF(E_DBG, L_SCAN, "File permissions changed: %s\n", path);
+
+      if (euidaccess(path, R_OK) < 0)
+	{
+	  DPRINTF(E_LOG, L_SCAN, "File access to '%s' failed: %s\n", path, strerror(errno));
+
+	  db_file_delete_bypath(path);;
+	}
+      else if (db_file_id_bypath(path) <= 0)
+	{
+	  DPRINTF(E_LOG, L_SCAN, "File access to '%s' achieved\n", path);
+
+	  ie->mask |= IN_CLOSE_WRITE;
+	}
     }
 
   if (ie->mask & IN_MOVED_TO)
     {
-      ret = db_file_enable_bycookie(ie->cookie, wi->path);
+      DPRINTF(E_DBG, L_SCAN, "File moved to: %s\n", path);
+
+      ret = db_file_enable_bycookie(ie->cookie, path);
 
       if (ret <= 0)
 	{
@@ -978,13 +1334,31 @@ process_inotify_file(struct watch_info *wi, char *path, struct inotify_event *ie
 	   * We want to scan the new file and we want to rescan the
 	   * playlist to update playlist items (relative items).
 	   */
-	  ie->mask |= IN_CREATE;
-	  db_pl_enable_bycookie(ie->cookie, wi->path);
+	  ie->mask |= IN_CLOSE_WRITE;
+	  db_pl_enable_bycookie(ie->cookie, path);
 	}
     }
 
-  if (ie->mask & (IN_MODIFY | IN_CREATE | IN_CLOSE_WRITE))
+  if (ie->mask & IN_CREATE)
     {
+      DPRINTF(E_DBG, L_SCAN, "File created: %s\n", path);
+
+      ret = lstat(path, &sb);
+      if (ret < 0)
+	{
+	  DPRINTF(E_LOG, L_SCAN, "Could not lstat() '%s': %s\n", path, strerror(errno));
+
+	  return;
+	}
+
+      if (S_ISFIFO(sb.st_mode))
+	ie->mask |= IN_CLOSE_WRITE;
+    }
+
+  if (ie->mask & IN_CLOSE_WRITE)
+    {
+      DPRINTF(E_DBG, L_SCAN, "File closed: %s\n", path);
+
       ret = lstat(path, &sb);
       if (ret < 0)
 	{
@@ -1023,9 +1397,18 @@ process_inotify_file(struct watch_info *wi, char *path, struct inotify_event *ie
 	    }
 	}
 
-      compilation = check_compilation(path);
+      type = 0;
+      if (check_speciallib(path, "compilations"))
+	type |= F_SCAN_TYPE_COMPILATION;
+      if (check_speciallib(path, "podcasts"))
+	type |= F_SCAN_TYPE_PODCAST;
+      if (check_speciallib(path, "audiobooks"))
+	type |= F_SCAN_TYPE_AUDIOBOOK;
 
-      process_file(file, sb.st_mtime, sb.st_size, compilation, 0);
+      if (S_ISREG(sb.st_mode))
+	process_file(file, sb.st_mtime, sb.st_size, F_SCAN_TYPE_FILE | type, 0);
+      else if (S_ISFIFO(sb.st_mode))
+	process_file(file, sb.st_mtime, sb.st_size, F_SCAN_TYPE_PIPE | type, 0);
 
       if (deref)
 	free(deref);
@@ -1327,6 +1710,46 @@ kqueue_cb(int fd, short event, void *arg)
 }
 #endif /* __FreeBSD__ || __FreeBSD_kernel__ */
 
+/* Thread: main & scan */
+static int
+inofd_event_set(void)
+{
+#if defined(__linux__)
+  inofd = inotify_init1(IN_CLOEXEC);
+  if (inofd < 0)
+    {
+      DPRINTF(E_FATAL, L_SCAN, "Could not create inotify fd: %s\n", strerror(errno));
+
+      return -1;
+    }
+
+  event_set(&inoev, inofd, EV_READ, inotify_cb, NULL);
+
+#elif defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
+
+  inofd = kqueue();
+  if (inofd < 0)
+    {
+      DPRINTF(E_FATAL, L_SCAN, "Could not create kqueue: %s\n", strerror(errno));
+
+      return -1;
+    }
+
+  event_set(&inoev, inofd, EV_READ, kqueue_cb, NULL);
+#endif
+
+  event_base_set(evbase_scan, &inoev);
+
+  return 0;
+}
+
+/* Thread: main & scan */
+static void
+inofd_event_unset(void)
+{
+  event_del(&inoev);
+  close(inofd);
+}
 
 /* Thread: scan */
 static void
@@ -1338,6 +1761,12 @@ exit_cb(int fd, short event, void *arg)
 }
 
 
+int
+filescanner_status(void)
+{
+  return scan_exit;
+}
+
 /* Thread: main */
 int
 filescanner_init(void)
@@ -1376,32 +1805,12 @@ filescanner_init(void)
     }
 #endif /* USE_EVENTFD */
 
-#if defined(__linux__)
-  inofd = inotify_init1(IN_CLOEXEC);
-  if (inofd < 0)
+  ret = inofd_event_set();
+  if (ret < 0)
     {
-      DPRINTF(E_FATAL, L_SCAN, "Could not create inotify fd: %s\n", strerror(errno));
-
       goto ino_fail;
     }
 
-  event_set(&inoev, inofd, EV_READ, inotify_cb, NULL);
-
-#elif defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
-
-  inofd = kqueue();
-  if (inofd < 0)
-    {
-      DPRINTF(E_FATAL, L_SCAN, "Could not create kqueue: %s\n", strerror(errno));
-
-      goto ino_fail;
-    }
-
-  event_set(&inoev, inofd, EV_READ, kqueue_cb, NULL);
-#endif
-
-  event_base_set(evbase_scan, &inoev);
-
 #ifdef USE_EVENTFD
   event_set(&exitev, exit_efd, EV_READ, exit_cb, NULL);
 #else
@@ -1469,7 +1878,7 @@ filescanner_deinit(void)
       return;
     }
 
-  event_del(&inoev);
+  inofd_event_unset();
 
 #ifdef USE_EVENTFD
   close(exit_efd);
@@ -1477,6 +1886,5 @@ filescanner_deinit(void)
   close(exit_pipe[0]);
   close(exit_pipe[1]);
 #endif
-  close(inofd);
   event_base_free(evbase_scan);
 }
