@@ -37,7 +37,11 @@
 #include <avl.h>
 #include <plist/plist.h>
 
-#include "evhttp/evhttp.h"
+#ifdef HAVE_LIBEVENT2
+# include <event2/http.h>
+#else
+# include "evhttp/evhttp.h"
+#endif
 
 #include "logger.h"
 #include "db.h"
@@ -260,20 +264,22 @@ check_meta(plist_t dict)
   return 0;
 }
 
-
-/* Best-effort attempt at locating the file:
- *  - first exact location given in itml
- *  - somewhere under the location of itml
- *  - anywhere (filename only)
- */
 static int
-find_track_file(char *location, char *base)
+find_track_file(char *location)
 {
-  char *filename;
+  int ret;
   int plen;
   int mfi_id;
+  char *entry;
+  char *ptr;
 
   location = evhttp_decode_uri(location);
+  if (!location)
+    {
+      DPRINTF(E_LOG, L_SCAN, "Could not decode iTunes XML playlist url.\n");
+
+      return 0;
+    }
 
   plen = strlen("file://localhost/");
 
@@ -281,58 +287,48 @@ find_track_file(char *location, char *base)
   if (strncmp(location, "file://localhost/", plen) != 0)
     return 0;
 
-  /* Windows pathspec, from iTunes Win32 */
-  if (location[plen + 1] == ':')
-    plen += 2;
-  else
-    plen -= 1;
-
-  /* Try exact path first */
-  filename = m_realpath(location + plen);
-  if (filename)
+  /* Now search for the library item where the path has closest match to playlist item */
+  /* Succes is when we find an unambiguous match, or when we no longer can expand the  */
+  /* the path to refine our search.                                                    */
+  entry = NULL;
+  do
     {
-      mfi_id = db_file_id_bypath(filename);
-
-      free(filename);
-
-      if (mfi_id > 0)
+      ptr = strrchr(location, '/');
+      if (entry)
+	*(entry - 1) = '/';
+      if (ptr)
 	{
-	  free(location);
-
-	  return mfi_id;
+	  *ptr = '\0';
+	  entry = ptr + 1;
 	}
-    }
+      else
+	entry = location;
 
-  filename = strrchr(location, '/');
-  if (!filename)
+      DPRINTF(E_SPAM, L_SCAN, "iTunes XML playlist entry is now %s\n", entry);
+      ret = db_files_get_count_bymatch(entry);
+
+    } while (ptr && (ret > 1));
+
+  if (ret > 0)
     {
-      DPRINTF(E_WARN, L_SCAN, "Could not extract filename from location\n");
+      mfi_id = db_file_id_bymatch(entry);
+      DPRINTF(E_DBG, L_SCAN, "Found iTunes XML playlist entry match, id is %d, entry is %s\n", mfi_id, entry);
+
+      free(location);
+      return mfi_id;
+    }
+  else
+    {
+      DPRINTF(E_DBG, L_SCAN, "No match for iTunes XML playlist entry %s\n", entry);
 
       free(location);
       return 0;
     }
 
-  filename++;
-
-  /* Try to locate the file under the playlist location */
-  mfi_id = db_file_id_byfilebase(filename, base);
-  if (mfi_id > 0)
-    {
-      free(location);
-
-      return mfi_id;
-    }
-
-  /* Last resort, filename only */
-  mfi_id = db_file_id_byfile(filename);
-
-  free(location);
-
-  return mfi_id;
 }
 
 static int
-process_track_file(plist_t trk, char *base)
+process_track_file(plist_t trk)
 {
   char *location;
   struct media_file_info *mfi;
@@ -354,7 +350,7 @@ process_track_file(plist_t trk, char *base)
       return 0;
     }
 
-  mfi_id = find_track_file(location, base);
+  mfi_id = find_track_file(location);
 
   if (mfi_id <= 0)
     {
@@ -427,6 +423,13 @@ process_track_file(plist_t trk, char *base)
 	}
     }
 
+  /* Set media_kind to 4 (Podcast) if Podcast is true */
+  ret = get_dictval_bool_from_key(trk, "Podcast", &boolean); 
+  if ((ret == 0) && boolean)
+    {
+      mfi->media_kind = 4;
+    }
+
   /* Don't let album_artist set to "Unknown artist" if we've
    * filled artist from the iTunes data in the meantime
    */
@@ -466,7 +469,7 @@ process_track_stream(plist_t trk)
 }
 
 static int
-process_tracks(plist_t tracks, char *base)
+process_tracks(plist_t tracks)
 {
   plist_t trk;
   plist_dict_iter iter;
@@ -537,7 +540,7 @@ process_tracks(plist_t tracks, char *base)
       if (strcmp(str, "URL") == 0)
 	mfi_id = process_track_stream(trk);
       else if (strcmp(str, "File") == 0)
-	mfi_id = process_track_file(trk, base);
+	mfi_id = process_track_file(trk);
       else
 	{
 	  DPRINTF(E_LOG, L_SCAN, "Unknown track type: %s\n", str);
@@ -853,7 +856,7 @@ scan_itunes_itml(char *file)
 
   *ptr = '\0';
 
-  ret = process_tracks(node, file);
+  ret = process_tracks(node);
   if (ret <= 0)
     {
       DPRINTF(E_WARN, L_SCAN, "No tracks loaded\n");
