@@ -42,8 +42,6 @@
 #endif
 
 #include <zlib.h>
-#include <event.h>
-#include "evhttp/evhttp.h"
 
 #include "logger.h"
 #include "db.h"
@@ -55,7 +53,6 @@
 #include "httpd_dacp.h"
 #include "transcode.h"
 
-
 /*
  * HTTP client quirks by User-Agent, from mt-daapd
  *
@@ -125,12 +122,19 @@ static struct event exitev;
 static struct evhttp *evhttpd;
 static pthread_t tid_httpd;
 
+#ifdef HAVE_LIBEVENT2_OLD
+struct stream_ctx *g_st;
+#endif
 
 static void
 stream_end(struct stream_ctx *st, int failed)
 {
-  if (st->req->evcon)
-    evhttp_connection_set_closecb(st->req->evcon, NULL, NULL);
+  struct evhttp_connection *evcon;
+
+  evcon = evhttp_request_get_connection(st->req);
+
+  if (evcon)
+    evhttp_connection_set_closecb(evcon, NULL, NULL);
 
   if (!failed)
     evhttp_send_reply_end(st->req);
@@ -145,6 +149,11 @@ stream_end(struct stream_ctx *st, int failed)
       close(st->fd);
     }
 
+#ifdef HAVE_LIBEVENT2_OLD
+  if (g_st == st)
+    g_st = NULL;
+#endif
+
   free(st);
 }
 
@@ -179,6 +188,15 @@ stream_chunk_resched_cb(struct evhttp_connection *evcon, void *arg)
     }
 }
 
+#ifdef HAVE_LIBEVENT2_OLD
+static void
+stream_chunk_resched_cb_wrapper(struct bufferevent *bufev, void *arg)
+{
+  if (g_st)
+    stream_chunk_resched_cb(NULL, g_st);
+}
+#endif
+
 static void
 stream_chunk_xcode_cb(int fd, short event, void *arg)
 {
@@ -226,7 +244,17 @@ stream_chunk_xcode_cb(int fd, short event, void *arg)
   else
     ret = xcoded;
 
+#ifdef HAVE_LIBEVENT2_OLD
+  evhttp_send_reply_chunk(st->req, st->evbuf);
+
+  struct evhttp_connection *evcon = evhttp_request_get_connection(st->req);
+  struct bufferevent *bufev = evhttp_connection_get_bufferevent(evcon);
+
+  g_st = st; // Can't pass st to callback so use global - limits libevent 2.0 to a single stream
+  bufev->writecb = stream_chunk_resched_cb_wrapper;
+#else
   evhttp_send_reply_chunk_with_cb(st->req, st->evbuf, stream_chunk_resched_cb, st);
+#endif
 
   st->offset += ret;
 
@@ -282,7 +310,17 @@ stream_chunk_raw_cb(int fd, short event, void *arg)
 
   evbuffer_add(st->evbuf, st->buf, ret);
 
+#ifdef HAVE_LIBEVENT2_OLD
+  evhttp_send_reply_chunk(st->req, st->evbuf);
+
+  struct evhttp_connection *evcon = evhttp_request_get_connection(st->req);
+  struct bufferevent *bufev = evhttp_connection_get_bufferevent(evcon);
+
+  g_st = st; // Can't pass st to callback so use global - limits libevent 2.0 to a single stream
+  bufev->writecb = stream_chunk_resched_cb_wrapper;
+#else
   evhttp_send_reply_chunk_with_cb(st->req, st->evbuf, stream_chunk_resched_cb, st);
+#endif
 
   st->offset += ret;
 
@@ -296,7 +334,7 @@ stream_fail_cb(struct evhttp_connection *evcon, void *arg)
 
   st = (struct stream_ctx *)arg;
 
-  DPRINTF(E_LOG, L_HTTPD, "Connection failed; stopping streaming of file ID %d\n", st->id);
+  DPRINTF(E_WARN, L_HTTPD, "Connection failed; stopping streaming of file ID %d\n", st->id);
 
   /* Stop streaming */
   event_del(&st->ev);
@@ -314,6 +352,9 @@ httpd_stream_file(struct evhttp_request *req, int id)
   void (*stream_cb)(int fd, short event, void *arg);
   struct stat sb;
   struct timeval tv;
+  struct evhttp_connection *evcon;
+  struct evkeyvalq *input_headers;
+  struct evkeyvalq *output_headers;
   const char *param;
   const char *param_end;
   char buf[64];
@@ -325,7 +366,10 @@ httpd_stream_file(struct evhttp_request *req, int id)
 
   offset = 0;
   end_offset = 0;
-  param = evhttp_find_header(req->input_headers, "Range");
+
+  input_headers = evhttp_request_get_input_headers(req);
+
+  param = evhttp_find_header(input_headers, "Range");
   if (param)
     {
       DPRINTF(E_DBG, L_HTTPD, "Found Range header: %s\n", param);
@@ -341,7 +385,7 @@ httpd_stream_file(struct evhttp_request *req, int id)
       else
 	{
 	  param_end = strchr(param, '-');
-	  if (param_end)
+	  if (param_end && (strlen(param_end) > 1))
 	    {
 	      ret = safe_atoi64(param_end + 1, &end_offset);
 	      if (ret < 0)
@@ -387,7 +431,9 @@ httpd_stream_file(struct evhttp_request *req, int id)
   memset(st, 0, sizeof(struct stream_ctx));
   st->fd = -1;
 
-  transcode = transcode_needed(req->input_headers, mfi->codectype);
+  transcode = transcode_needed(input_headers, mfi->codectype);
+
+  output_headers = evhttp_request_get_output_headers(req);
 
   if (transcode)
     {
@@ -395,8 +441,8 @@ httpd_stream_file(struct evhttp_request *req, int id)
 
       stream_cb = stream_chunk_xcode_cb;
 
-      st->xcode = transcode_setup(mfi, &st->size, 1);
-      if (!st->xcode)
+      ret = transcode_setup(&st->xcode, mfi, &st->size, 1);
+      if (ret < 0)
 	{
 	  DPRINTF(E_WARN, L_HTTPD, "Transcoding setup failed, aborting streaming\n");
 
@@ -405,8 +451,8 @@ httpd_stream_file(struct evhttp_request *req, int id)
 	  goto out_free_st;
 	}
 
-      if (!evhttp_find_header(req->output_headers, "Content-Type"))
-	evhttp_add_header(req->output_headers, "Content-Type", "audio/wav");
+      if (!evhttp_find_header(output_headers, "Content-Type"))
+	evhttp_add_header(output_headers, "Content-Type", "audio/wav");
     }
   else
     {
@@ -470,21 +516,21 @@ httpd_stream_file(struct evhttp_request *req, int id)
 	    DPRINTF(E_LOG, L_HTTPD, "Content-Type too large for buffer, dropping\n");
 	  else
 	    {
-	      evhttp_remove_header(req->output_headers, "Content-Type");
-	      evhttp_add_header(req->output_headers, "Content-Type", buf);
+	      evhttp_remove_header(output_headers, "Content-Type");
+	      evhttp_add_header(output_headers, "Content-Type", buf);
 	    }
 	}
       /* If no Content-Type has been set and we're streaming audio, add a proper
        * Content-Type for the file we're streaming. Remember DAAP streams audio
        * with application/x-dmap-tagged as the Content-Type (ugh!).
        */
-      else if (!evhttp_find_header(req->output_headers, "Content-Type") && mfi->type)
+      else if (!evhttp_find_header(output_headers, "Content-Type") && mfi->type)
 	{
 	  ret = snprintf(buf, sizeof(buf), "audio/%s", mfi->type);
 	  if ((ret < 0) || (ret >= sizeof(buf)))
 	    DPRINTF(E_LOG, L_HTTPD, "Content-Type too large for buffer, dropping\n");
 	  else
-	    evhttp_add_header(req->output_headers, "Content-Type", buf);
+	    evhttp_add_header(output_headers, "Content-Type", buf);
 	}
     }
 
@@ -493,7 +539,7 @@ httpd_stream_file(struct evhttp_request *req, int id)
     {
       DPRINTF(E_LOG, L_HTTPD, "Could not allocate an evbuffer for streaming\n");
 
-      evhttp_clear_headers(req->output_headers);
+      evhttp_clear_headers(output_headers);
       evhttp_send_error(req, HTTP_SERVUNAVAIL, "Internal Server Error");
 
       goto out_cleanup;
@@ -504,7 +550,7 @@ httpd_stream_file(struct evhttp_request *req, int id)
     {
       DPRINTF(E_LOG, L_HTTPD, "Could not expand evbuffer for streaming\n");
 
-      evhttp_clear_headers(req->output_headers);
+      evhttp_clear_headers(output_headers);
       evhttp_send_error(req, HTTP_SERVUNAVAIL, "Internal Server Error");
 
       goto out_cleanup;
@@ -518,7 +564,7 @@ httpd_stream_file(struct evhttp_request *req, int id)
     {
       DPRINTF(E_LOG, L_HTTPD, "Could not add one-shot event for streaming\n");
 
-      evhttp_clear_headers(req->output_headers);
+      evhttp_clear_headers(output_headers);
       evhttp_send_error(req, HTTP_SERVUNAVAIL, "Internal Server Error");
 
       goto out_cleanup;
@@ -541,7 +587,7 @@ httpd_stream_file(struct evhttp_request *req, int id)
 	  if ((ret < 0) || (ret >= sizeof(buf)))
 	    DPRINTF(E_LOG, L_HTTPD, "Content-Length too large for buffer, dropping\n");
 	  else
-	    evhttp_add_header(req->output_headers, "Content-Length", buf);
+	    evhttp_add_header(output_headers, "Content-Length", buf);
 	}
 
       evhttp_send_reply_start(req, HTTP_OK, "OK");
@@ -560,13 +606,13 @@ httpd_stream_file(struct evhttp_request *req, int id)
       if ((ret < 0) || (ret >= sizeof(buf)))
 	DPRINTF(E_LOG, L_HTTPD, "Content-Range too large for buffer, dropping\n");
       else
-	evhttp_add_header(req->output_headers, "Content-Range", buf);
+	evhttp_add_header(output_headers, "Content-Range", buf);
 
       ret = snprintf(buf, sizeof(buf), "%" PRIi64, ((end_offset) ? end_offset + 1 : (int64_t)st->size) - offset);
       if ((ret < 0) || (ret >= sizeof(buf)))
 	DPRINTF(E_LOG, L_HTTPD, "Content-Length too large for buffer, dropping\n");
       else
-	evhttp_add_header(req->output_headers, "Content-Length", buf);
+	evhttp_add_header(output_headers, "Content-Length", buf);
 
       evhttp_send_reply_start(req, 206, "Partial Content");
     }
@@ -581,7 +627,9 @@ httpd_stream_file(struct evhttp_request *req, int id)
     }
 #endif
 
-  evhttp_connection_set_closecb(req->evcon, stream_fail_cb, st);
+  evcon = evhttp_request_get_connection(req);
+
+  evhttp_connection_set_closecb(evcon, stream_fail_cb, st);
 
   DPRINTF(E_INFO, L_HTTPD, "Kicking off streaming for %s\n", mfi->path);
 
@@ -611,6 +659,7 @@ httpd_send_reply(struct evhttp_request *req, int code, const char *reason, struc
   unsigned char outbuf[128 * 1024];
   z_stream strm;
   struct evbuffer *gzbuf;
+  struct evkeyvalq *headers;
   const char *param;
   int flush;
   int zret;
@@ -623,7 +672,9 @@ httpd_send_reply(struct evhttp_request *req, int code, const char *reason, struc
       goto no_gzip;
     }
 
-  param = evhttp_find_header(req->input_headers, "Accept-Encoding");
+  headers = evhttp_request_get_input_headers(req);
+
+  param = evhttp_find_header(headers, "Accept-Encoding");
   if (!param)
     {
       DPRINTF(E_DBG, L_HTTPD, "Not gzipping; no Accept-Encoding header\n");
@@ -704,7 +755,9 @@ httpd_send_reply(struct evhttp_request *req, int code, const char *reason, struc
 
   deflateEnd(&strm);
 
-  evhttp_add_header(req->output_headers, "Content-Encoding", "gzip");
+  headers = evhttp_request_get_output_headers(req);
+
+  evhttp_add_header(headers, "Content-Encoding", "gzip");
   evhttp_send_reply(req, code, reason, gzbuf);
 
   evbuffer_free(gzbuf);
@@ -733,6 +786,7 @@ path_is_legal(char *path)
 static void
 redirect_to_index(struct evhttp_request *req, char *uri)
 {
+  struct evkeyvalq *headers;
   char buf[256];
   int slashed;
   int ret;
@@ -748,7 +802,9 @@ redirect_to_index(struct evhttp_request *req, char *uri)
       return;
     }
 
-  evhttp_add_header(req->output_headers, "Location", buf);
+  headers = evhttp_request_get_output_headers(req);
+
+  evhttp_add_header(headers, "Location", buf);
   evhttp_send_reply(req, HTTP_MOVETEMP, "Moved", NULL);
 }
 
@@ -756,12 +812,14 @@ redirect_to_index(struct evhttp_request *req, char *uri)
 static void
 serve_file(struct evhttp_request *req, char *uri)
 {
+  const char *host;
   char *ext;
   char path[PATH_MAX];
   char *deref;
   char *ctype;
   char *passwd;
   struct evbuffer *evbuf;
+  struct evkeyvalq *headers;
   struct stat sb;
   int fd;
   int i;
@@ -781,8 +839,9 @@ serve_file(struct evhttp_request *req, char *uri)
     }
   else
     {
-      if ((strcmp(req->remote_host, "::1") != 0)
-	  && (strcmp(req->remote_host, "127.0.0.1") != 0))
+      host = evhttp_request_get_host(req);
+      if ((strcmp(host, "::1") != 0)
+	  && (strcmp(host, "127.0.0.1") != 0))
 	{
 	  DPRINTF(E_LOG, L_HTTPD, "Remote web interface request denied; no password set\n");
 
@@ -912,8 +971,9 @@ serve_file(struct evhttp_request *req, char *uri)
 	}
     }
 
-  evhttp_add_header(req->output_headers, "Content-Type", ctype);
+  headers = evhttp_request_get_output_headers(req);
 
+  evhttp_add_header(headers, "Content-Type", ctype);
   evhttp_send_reply(req, HTTP_OK, "OK", evbuf);
 
   evbuffer_free(evbuf);
@@ -927,7 +987,7 @@ httpd_gen_cb(struct evhttp_request *req, void *arg)
   char *uri;
   char *ptr;
 
-  req_uri = evhttp_request_uri(req);
+  req_uri = evhttp_request_get_uri(req);
   if (!req_uri)
     {
       redirect_to_index(req, "/");
@@ -1013,6 +1073,7 @@ exit_cb(int fd, short event, void *arg)
 char *
 httpd_fixup_uri(struct evhttp_request *req)
 {
+  struct evkeyvalq *headers;
   const char *ua;
   const char *uri;
   const char *u;
@@ -1021,7 +1082,7 @@ httpd_fixup_uri(struct evhttp_request *req)
   char *f;
   int len;
 
-  uri = evhttp_request_uri(req);
+  uri = evhttp_request_get_uri(req);
   if (!uri)
     return NULL;
 
@@ -1030,7 +1091,8 @@ httpd_fixup_uri(struct evhttp_request *req)
   if (!q)
     return strdup(uri);
 
-  ua = evhttp_find_header(req->input_headers, "User-Agent");
+  headers = evhttp_request_get_input_headers(req);
+  ua = evhttp_find_header(headers, "User-Agent");
   if (!ua)
     return strdup(uri);
 
@@ -1095,6 +1157,7 @@ int
 httpd_basic_auth(struct evhttp_request *req, char *user, char *passwd, char *realm)
 {
   struct evbuffer *evbuf;
+  struct evkeyvalq *headers;
   char *header;
   const char *auth;
   char *authuser;
@@ -1102,7 +1165,8 @@ httpd_basic_auth(struct evhttp_request *req, char *user, char *passwd, char *rea
   int len;
   int ret;
 
-  auth = evhttp_find_header(req->input_headers, "Authorization");
+  headers = evhttp_request_get_input_headers(req);
+  auth = evhttp_find_header(headers, "Authorization");
   if (!auth)
     {
       DPRINTF(E_DBG, L_HTTPD, "No Authorization header\n");
@@ -1185,7 +1249,8 @@ httpd_basic_auth(struct evhttp_request *req, char *user, char *passwd, char *rea
       return -1;
     }
 
-  evhttp_add_header(req->output_headers, "WWW-Authenticate", header);
+  headers = evhttp_request_get_output_headers(req);
+  evhttp_add_header(headers, "WWW-Authenticate", header);
   evbuffer_add(evbuf, http_reply_401, strlen(http_reply_401));
   evhttp_send_reply(req, 401, "Unauthorized", evbuf);
 
