@@ -37,9 +37,6 @@
 # include <sys/eventfd.h>
 #endif
 
-#include <event.h>
-#include "evhttp/evhttp.h"
-
 #include "logger.h"
 #include "misc.h"
 #include "conffile.h"
@@ -50,7 +47,6 @@
 #include "db.h"
 #include "player.h"
 
-
 /* httpd event base, from httpd.c */
 extern struct event_base *evbase_httpd;
 
@@ -103,6 +99,23 @@ dacp_propget_nowplaying(struct evbuffer *evbuf, struct player_status *status, st
 static void
 dacp_propget_playingtime(struct evbuffer *evbuf, struct player_status *status, struct media_file_info *mfi);
 
+static void
+dacp_propget_fullscreenenabled(struct evbuffer *evbuf, struct player_status *status, struct media_file_info *mfi);
+static void
+dacp_propget_fullscreen(struct evbuffer *evbuf, struct player_status *status, struct media_file_info *mfi);
+static void
+dacp_propget_visualizerenabled(struct evbuffer *evbuf, struct player_status *status, struct media_file_info *mfi);
+static void
+dacp_propget_visualizer(struct evbuffer *evbuf, struct player_status *status, struct media_file_info *mfi);
+static void
+dacp_propget_itms_songid(struct evbuffer *evbuf, struct player_status *status, struct media_file_info *mfi);
+static void
+dacp_propget_haschapterdata(struct evbuffer *evbuf, struct player_status *status, struct media_file_info *mfi);
+static void
+dacp_propget_mediakind(struct evbuffer *evbuf, struct player_status *status, struct media_file_info *mfi);
+static void
+dacp_propget_extendedmediakind(struct evbuffer *evbuf, struct player_status *status, struct media_file_info *mfi);
+
 /* Forward - properties setters */
 static void
 dacp_propset_volume(const char *value, struct evkeyvalq *query);
@@ -141,31 +154,14 @@ static int seek_target;
 static void
 dacp_nowplaying(struct evbuffer *evbuf, struct player_status *status, struct media_file_info *mfi)
 {
-  char canp[16];
-
   if ((status->status == PLAY_STOPPED) || !mfi)
     return;
 
-  memset(canp, 0, sizeof(canp));
-
-  canp[3] = 1; /* 0-3 database ID */
-
-  canp[4]  = (status->plid >> 24) & 0xff;
-  canp[5]  = (status->plid >> 16) & 0xff;
-  canp[6]  = (status->plid >> 8) & 0xff;
-  canp[7]  = status->plid & 0xff;
-
-  canp[8]  = (status->pos_pl >> 24) & 0xff; /* 8-11 position in playlist */
-  canp[9]  = (status->pos_pl >> 16) & 0xff;
-  canp[10] = (status->pos_pl >> 8) & 0xff;
-  canp[11] = status->pos_pl & 0xff;
-
-  canp[12] = (status->id >> 24) & 0xff; /* 12-15 track ID */
-  canp[13] = (status->id >> 16) & 0xff;
-  canp[14] = (status->id >> 8) & 0xff;
-  canp[15] = status->id & 0xff;
-
-  dmap_add_literal(evbuf, "canp", canp, sizeof(canp));
+  dmap_add_container(evbuf, "canp", 16);
+  dmap_add_raw_uint32(evbuf, 1); /* Database */
+  dmap_add_raw_uint32(evbuf, status->plid);
+  dmap_add_raw_uint32(evbuf, status->pos_pl);
+  dmap_add_raw_uint32(evbuf, status->id);
 
   dmap_add_string(evbuf, "cann", mfi->title);
   dmap_add_string(evbuf, "cana", mfi->artist);
@@ -223,22 +219,34 @@ make_playstatusupdate(struct evbuffer *evbuf)
 
   dmap_add_int(psu, "cmsr", current_rev); /* 12 */
 
-  dmap_add_char(psu, "cavc", 1);              /* 9 */ /* volume controllable */
-  dmap_add_char(psu, "caps", status.status);  /* 9 */ /* play status, 2 = stopped, 3 = paused, 4 = playing */
-  dmap_add_char(psu, "cash", status.shuffle); /* 9 */ /* shuffle, true/false */
-  dmap_add_char(psu, "carp", status.repeat);  /* 9 */ /* repeat, 0 = off, 1 = repeat song, 2 = repeat (playlist) */
-
-  dmap_add_int(psu, "caas", 2);           /* 12 */ /* available shuffle states */
-  dmap_add_int(psu, "caar", 6);           /* 12 */ /* available repeat states */
+  dmap_add_char(psu, "caps", status.status);  /*  9 */ /* play status, 2 = stopped, 3 = paused, 4 = playing */
+  dmap_add_char(psu, "cash", status.shuffle); /*  9 */ /* shuffle, true/false */
+  dmap_add_char(psu, "carp", status.repeat);  /*  9 */ /* repeat, 0 = off, 1 = repeat song, 2 = repeat (playlist) */
+  dmap_add_char(psu, "cafs", 0);              /*  9 */ /* dacp.fullscreen */
+  dmap_add_char(psu, "cavs", 0);              /*  9 */ /* dacp.visualizer */
+  dmap_add_char(psu, "cavc", 1);              /*  9 */ /* volume controllable */
+  dmap_add_int(psu, "caas", 2);               /* 12 */ /* available shuffle states */
+  dmap_add_int(psu, "caar", 6);               /* 12 */ /* available repeat states */
+  dmap_add_char(psu, "cafe", 0);              /*  9 */ /* dacp.fullscreenenabled */
+  dmap_add_char(psu, "cave", 0);              /*  9 */ /* dacp.visualizerenabled */
 
   if (mfi)
     {
       dacp_nowplaying(psu, &status, mfi);
+
+      dmap_add_int(psu, "casa", 1);           /* 12 */ /* unknown */
+      dmap_add_int(psu, "astm", mfi->song_length);
+      dmap_add_char(psu, "casc", 1);         /* Maybe an indication of extra data? */
+      dmap_add_char(psu, "caks", 6);         /* Unknown */
+
       dacp_playingtime(psu, &status, mfi);
 
       free_mfi(mfi, 0);
     }
 
+  dmap_add_char(psu, "casu", 1);              /*  9 */ /* unknown */
+  dmap_add_char(psu, "ceQu", 0);              /*  9 */ /* unknown */
+
   dmap_add_container(evbuf, "cmst", EVBUFFER_LENGTH(psu));    /* 8 + len */
 
   ret = evbuffer_add_buffer(evbuf, psu);
@@ -259,6 +267,7 @@ playstatusupdate_cb(int fd, short what, void *arg)
   struct dacp_update_request *ur;
   struct evbuffer *evbuf;
   struct evbuffer *update;
+  struct evhttp_connection *evcon;
   int ret;
 
 #ifdef USE_EVENTFD
@@ -304,7 +313,9 @@ playstatusupdate_cb(int fd, short what, void *arg)
     {
       update_requests = ur->next;
 
-      evhttp_connection_set_closecb(ur->req->evcon, NULL, NULL);
+      evcon = evhttp_request_get_connection(ur->req);
+      if (evcon)
+	evhttp_connection_set_closecb(evcon, NULL, NULL);
 
       evbuffer_add(evbuf, EVBUFFER_DATA(update), EVBUFFER_LENGTH(update));
 
@@ -349,13 +360,15 @@ update_fail_cb(struct evhttp_connection *evcon, void *arg)
 {
   struct dacp_update_request *ur;
   struct dacp_update_request *p;
+  struct evhttp_connection *evc;
 
   ur = (struct dacp_update_request *)arg;
 
   DPRINTF(E_DBG, L_DACP, "Update request: client closed connection\n");
 
-  if (ur->req->evcon)
-    evhttp_connection_set_closecb(ur->req->evcon, NULL, NULL);
+  evc = evhttp_request_get_connection(ur->req);
+  if (evc)
+    evhttp_connection_set_closecb(evc, NULL, NULL);
 
   if (ur == update_requests)
     update_requests = ur->next;
@@ -432,6 +445,54 @@ dacp_propget_playingtime(struct evbuffer *evbuf, struct player_status *status, s
   dacp_playingtime(evbuf, status, mfi);
 }
 
+static void
+dacp_propget_fullscreenenabled(struct evbuffer *evbuf, struct player_status *status, struct media_file_info *mfi)
+{
+	// TODO
+}
+
+static void
+dacp_propget_fullscreen(struct evbuffer *evbuf, struct player_status *status, struct media_file_info *mfi)
+{
+	// TODO
+}
+
+static void
+dacp_propget_visualizerenabled(struct evbuffer *evbuf, struct player_status *status, struct media_file_info *mfi)
+{
+	// TODO
+}
+
+static void
+dacp_propget_visualizer(struct evbuffer *evbuf, struct player_status *status, struct media_file_info *mfi)
+{
+	// TODO
+}
+
+static void
+dacp_propget_itms_songid(struct evbuffer *evbuf, struct player_status *status, struct media_file_info *mfi)
+{
+	// TODO
+}
+
+static void
+dacp_propget_haschapterdata(struct evbuffer *evbuf, struct player_status *status, struct media_file_info *mfi)
+{
+	// TODO
+}
+
+static void
+dacp_propget_mediakind(struct evbuffer *evbuf, struct player_status *status, struct media_file_info *mfi)
+{
+	// TODO
+}
+
+static void
+dacp_propget_extendedmediakind(struct evbuffer *evbuf, struct player_status *status, struct media_file_info *mfi)
+{
+	// TODO
+}
+
 /* Properties setters */
 static void
 dacp_propset_volume(const char *value, struct evkeyvalq *query)
@@ -624,27 +685,30 @@ dacp_propset_userrating(const char *value, struct evkeyvalq *query)
 static void
 dacp_reply_ctrlint(struct evhttp_request *req, struct evbuffer *evbuf, char **uri, struct evkeyvalq *query)
 {
-  dmap_add_container(evbuf, "caci", 127); /* 8 + len */
-  dmap_add_int(evbuf, "mstt", 200);       /* 12 */
-  dmap_add_char(evbuf, "muty", 0);        /* 9 */
-  dmap_add_int(evbuf, "mtco", 1);         /* 12 */
-  dmap_add_int(evbuf, "mrco", 1);         /* 12 */
-  dmap_add_container(evbuf, "mlcl", 125); /* 8 + len */
-  dmap_add_container(evbuf, "mlit", 117); /* 8 + len */
-  dmap_add_int(evbuf, "miid", 1);         /* 12 */ /* Database ID */
-  dmap_add_char(evbuf, "cmik", 1);        /* 9 */
+  /* /ctrl-int */
+  /* If tags are added or removed container sizes should be adjusted too */
+  dmap_add_container(evbuf, "caci", 194); /*  8, unknown dacp container - size of content */
+  dmap_add_int(evbuf, "mstt", 200);       /* 12, dmap.status */
+  dmap_add_char(evbuf, "muty", 0);        /*  9, dmap.updatetype */
+  dmap_add_int(evbuf, "mtco", 1);         /* 12, dmap.specifiedtotalcount */
+  dmap_add_int(evbuf, "mrco", 1);         /* 12, dmap.returnedcount */
+  dmap_add_container(evbuf, "mlcl", 141); /*  8, dmap.listing - size of content */
+  dmap_add_container(evbuf, "mlit", 133); /*  8, dmap.listingitem - size of content */
+  dmap_add_int(evbuf, "miid", 1);         /* 12, dmap.itemid - database ID */
+  dmap_add_char(evbuf, "cmik", 1);        /*  9, unknown */
 
-  dmap_add_int(evbuf, "cmpr", (2 << 16 | 1)); /* 12 */
-  dmap_add_int(evbuf, "capr", (2 << 16 | 2)); /* 12 */
+  dmap_add_int(evbuf, "cmpr", (2 << 16 | 2)); /* 12, dmcp.protocolversion */
+  dmap_add_int(evbuf, "capr", (2 << 16 | 5)); /* 12, dacp.protocolversion */
 
-  dmap_add_char(evbuf, "cmsp", 1);        /* 9 */
-  dmap_add_char(evbuf, "aeFR", 0x64);     /* 9 */
-  dmap_add_char(evbuf, "cmsv", 1);        /* 9 */
-  dmap_add_char(evbuf, "cass", 1);        /* 9 */
-  dmap_add_char(evbuf, "caov", 1);        /* 9 */
-  dmap_add_char(evbuf, "casu", 1);        /* 9 */
-  dmap_add_char(evbuf, "ceSG", 1);        /* 9 */
-  dmap_add_char(evbuf, "cmrl", 1);        /* 9 */
+  dmap_add_char(evbuf, "cmsp", 1);        /*  9, unknown */
+  dmap_add_char(evbuf, "aeFR", 0x64);     /*  9, unknown */
+  dmap_add_char(evbuf, "cmsv", 1);        /*  9, unknown */
+  dmap_add_char(evbuf, "cass", 1);        /*  9, unknown */
+  dmap_add_char(evbuf, "caov", 1);        /*  9, unknown */
+  dmap_add_char(evbuf, "casu", 1);        /*  9, unknown */
+  dmap_add_char(evbuf, "ceSG", 1);        /*  9, unknown */
+  dmap_add_char(evbuf, "cmrl", 1);        /*  9, unknown */
+  dmap_add_long(evbuf, "ceSX", (1 << 1 | 1));  /* 16, unknown dacp - lowest bit announces support for playqueue-contents/-edit */
 
   httpd_send_reply(req, HTTP_OK, "OK", evbuf);
 }
@@ -682,8 +746,8 @@ dacp_reply_cue_play(struct evhttp_request *req, struct evbuffer *evbuf, char **u
     {
       sort = evhttp_find_header(query, "sort");
 
-      ps = player_queue_make_daap(cuequery, sort);
-      if (!ps)
+      ret = player_queue_make_daap(&ps, cuequery, NULL, sort, 0);
+      if (ret < 0)
 	{
 	  DPRINTF(E_LOG, L_DACP, "Could not build song queue\n");
 
@@ -714,6 +778,10 @@ dacp_reply_cue_play(struct evhttp_request *req, struct evbuffer *evbuf, char **u
 	DPRINTF(E_LOG, L_DACP, "Invalid index (%s) in cue request\n", param);
     }
 
+  /* If selection was from Up Next queue (command will be playnow), then index is relative */
+  if ((param = evhttp_find_header(query, "command")) && (strcmp(param, "playnow") == 0))
+    id += status.pos_pl;
+
   ret = player_playback_start(&id);
   if (ret < 0)
     {
@@ -791,7 +859,9 @@ dacp_reply_playspec(struct evhttp_request *req, struct evbuffer *evbuf, char **u
   int ret;
 
   /* /ctrl-int/1/playspec?database-spec='dmap.persistentid:0x1'&container-spec='dmap.persistentid:0x5'&container-item-spec='dmap.containeritemid:0x9'
-   * With our DAAP implementation, container-spec is the playlist ID and container-item-spec is the song ID
+   * or (Apple Remote when playing a Podcast)
+   * /ctrl-int/1/playspec?database-spec='dmap.persistentid:0x1'&container-spec='dmap.persistentid:0x5'&item-spec='dmap.itemid:0x9'
+   * With our DAAP implementation, container-spec is the playlist ID and container-item-spec/item-spec is the song ID
    */
 
   s = daap_session_find(req, query, evbuf);
@@ -830,10 +900,11 @@ dacp_reply_playspec(struct evhttp_request *req, struct evbuffer *evbuf, char **u
   if (!shuffle)
     {
       /* Start song ID */
-      param = evhttp_find_header(query, "container-item-spec");
-      if (!param)
+      if ((param = evhttp_find_header(query, "item-spec")))
+	plid = 0; // This is a podcast/audiobook - just play a single item, not a playlist
+      else if (!(param = evhttp_find_header(query, "container-item-spec")))
 	{
-	  DPRINTF(E_LOG, L_DACP, "No container-item-spec in playspec request\n");
+	  DPRINTF(E_LOG, L_DACP, "No container-item-spec/item-spec in playspec request\n");
 
 	  goto out_fail;
 	}
@@ -841,7 +912,7 @@ dacp_reply_playspec(struct evhttp_request *req, struct evbuffer *evbuf, char **u
       param = strchr(param, ':');
       if (!param)
 	{
-	  DPRINTF(E_LOG, L_DACP, "Malformed container-item-spec parameter in playspec request\n");
+	  DPRINTF(E_LOG, L_DACP, "Malformed container-item-spec/item-spec parameter in playspec request\n");
 
 	  goto out_fail;
 	}
@@ -850,7 +921,7 @@ dacp_reply_playspec(struct evhttp_request *req, struct evbuffer *evbuf, char **u
       ret = safe_hextou32(param, &id);
       if (ret < 0)
 	{
-	  DPRINTF(E_LOG, L_DACP, "Couldn't convert container-item-spec to an integer in playspec (%s)\n", param);
+	  DPRINTF(E_LOG, L_DACP, "Couldn't convert container-item-spec/item-spec to an integer in playspec (%s)\n", param);
 
 	  goto out_fail;
 	}
@@ -917,19 +988,29 @@ static void
 dacp_reply_playpause(struct evhttp_request *req, struct evbuffer *evbuf, char **uri, struct evkeyvalq *query)
 {
   struct daap_session *s;
+  struct player_status status;
   int ret;
 
   s = daap_session_find(req, query, evbuf);
   if (!s)
     return;
 
-  ret = player_playback_start(NULL);
-  if (ret < 0)
-    {
-      DPRINTF(E_LOG, L_DACP, "Player returned an error for start after pause\n");
 
-      evhttp_send_error(req, 500, "Internal Server Error");
-      return;
+  player_get_status(&status);
+  if (status.status == PLAY_PLAYING)
+    {
+      player_playback_pause();
+    }
+  else
+    {
+      ret = player_playback_start(NULL);
+      if (ret < 0)
+	{
+	  DPRINTF(E_LOG, L_DACP, "Player returned an error for start after pause\n");
+
+	  evhttp_send_error(req, 500, "Internal Server Error");
+	  return;
+        }
     }
 
   /* 204 No Content is the canonical reply */
@@ -1045,11 +1126,563 @@ dacp_reply_playresume(struct evhttp_request *req, struct evbuffer *evbuf, char *
   evhttp_send_reply(req, HTTP_NOCONTENT, "No Content", evbuf);
 }
 
+static struct player_source *
+next_ps(struct player_source *ps, char shuffle)
+{
+  if (shuffle)
+    return ps->shuffle_next;
+  else
+    return ps->pl_next;
+}
+
+static int
+playqueuecontents_add_source(struct evbuffer *songlist, uint32_t source_id, int pos_in_queue, uint32_t plid)
+{
+  struct evbuffer *song;
+  struct media_file_info *mfi;
+  int ret;
+
+  song = evbuffer_new();
+  if (!song)
+    {
+      DPRINTF(E_LOG, L_DACP, "Could not allocate song evbuffer for playqueue-contents\n");
+      return -1;
+    }
+
+  mfi = db_file_fetch_byid(source_id);
+  if (!mfi)
+    {
+      DPRINTF(E_LOG, L_DACP, "Could not fetch file id %d\n", source_id);
+      return -1;
+    }
+  dmap_add_container(song, "ceQs", 16);
+  dmap_add_raw_uint32(song, 1); /* Database */
+  dmap_add_raw_uint32(song, plid);
+  dmap_add_raw_uint32(song, 0); /* Should perhaps be playlist index? */
+  dmap_add_raw_uint32(song, mfi->id);
+  dmap_add_string(song, "ceQn", mfi->title);
+  dmap_add_string(song, "ceQr", mfi->artist);
+  dmap_add_string(song, "ceQa", mfi->album);
+  dmap_add_string(song, "ceQg", mfi->genre);
+  dmap_add_long(song, "asai", mfi->songalbumid);
+  dmap_add_int(song, "cmmk", mfi->media_kind);
+  dmap_add_int(song, "casa", 1); /* Unknown  */
+  dmap_add_int(song, "astm", mfi->song_length);
+  dmap_add_char(song, "casc", 1); /* Maybe an indication of extra data? */
+  dmap_add_char(song, "caks", 6); /* Unknown */
+  dmap_add_int(song, "ceQI", pos_in_queue);
+
+  dmap_add_container(songlist, "mlit", EVBUFFER_LENGTH(song));
+
+  ret = evbuffer_add_buffer(songlist, song);
+  evbuffer_free(song);
+  free_mfi(mfi, 0);
+
+  if (ret < 0)
+    {
+      DPRINTF(E_LOG, L_DACP, "Could not add song to songlist for playqueue-contents\n");
+      return ret;
+    }
+
+  return 0;
+}
+
+static void
+dacp_reply_playqueuecontents(struct evhttp_request *req, struct evbuffer *evbuf, char **uri,
+    struct evkeyvalq *query)
+{
+  struct daap_session *s;
+  struct evbuffer *songlist;
+  struct evbuffer *playlists;
+  struct player_source *ps;
+  struct player_source *head;
+  struct player_status status;
+  struct player_history *history;
+  const char *param;
+  int span;
+  int i;
+  int n;
+  int songlist_length;
+  int ret;
+  int start_index;
+
+  /* /ctrl-int/1/playqueue-contents?span=50&session-id=... */
+
+  s = daap_session_find(req, query, evbuf);
+  if (!s)
+    return;
+
+  DPRINTF(E_DBG, L_DACP, "Fetching playqueue contents\n");
+
+  span = 50; /* Default */
+  param = evhttp_find_header(query, "span");
+  if (param)
+    {
+      ret = safe_atoi32(param, &span);
+      if (ret < 0)
+	DPRINTF(E_LOG, L_DACP, "Invalid span value in playqueue-contents request\n");
+    }
+
+  i = 0;
+  n = 0; // count of songs in songlist
+  songlist = evbuffer_new();
+  if (!songlist)
+    {
+      DPRINTF(E_LOG, L_DACP, "Could not allocate songlist evbuffer for playqueue-contents\n");
+
+      dmap_send_error(req, "ceQR", "Out of memory");
+      return;
+    }
+
+  /*
+   * If the span parameter is negativ make song list for Previously Played,
+   * otherwise make song list for Up Next and begin with first song after playlist position.
+   */
+  if (span < 0)
+    {
+      history = player_history_get();
+      if (abs(span) > history->count)
+	{
+	  start_index = history->start_index;
+	}
+      else
+	{
+	  start_index = (history->start_index + history->count - abs(span)) % MAX_HISTORY_COUNT;
+	}
+      for (n = 0; n < history->count && n < abs(span); n++)
+	{
+	  ret = playqueuecontents_add_source(songlist, history->id[(start_index + n) % MAX_HISTORY_COUNT], (n + 1), status.plid);
+	  if (ret < 0)
+	    {
+	      DPRINTF(E_LOG, L_DACP, "Could not add song to songlist for playqueue-contents\n");
+
+	      dmap_send_error(req, "ceQR", "Out of memory");
+	      return;
+	    }
+	}
+    }
+  else
+    {
+      player_get_status(&status);
+
+      /* Get queue and make songlist only if playing or paused */
+      if (status.status != PLAY_STOPPED)
+	{
+	  /* Fast forward to song currently being played */
+	  head = player_queue_get();
+	  if (head)
+	    {
+	      ps = head;
+	      while ((ps->id != status.id) && (ps = next_ps(ps, status.shuffle)) && (ps != head))
+		i++;
+
+	      while ((n < abs(span)) && (ps = next_ps(ps, status.shuffle)) && (ps != head))
+		{
+		  n++;
+
+		  ret = playqueuecontents_add_source(songlist, ps->id, (n + i + 1), status.plid);
+		  if (ret < 0)
+		    {
+		      DPRINTF(E_LOG, L_DACP, "Could not add song to songlist for playqueue-contents\n");
+
+		      dmap_send_error(req, "ceQR", "Out of memory");
+		      return;
+		    }
+		}
+	    }
+	}
+    }
+
+  /* Playlists are hist, curr and main. */
+  playlists = evbuffer_new();
+  if (!playlists)
+    {
+      DPRINTF(E_LOG, L_DACP, "Could not allocate playlists evbuffer for playqueue-contents\n");
+      if (songlist)
+	evbuffer_free(songlist);
+
+      dmap_send_error(req, "ceQR", "Out of memory");
+      return;
+    }
+
+  dmap_add_container(playlists, "mlit", 61);
+  dmap_add_string(playlists, "ceQk", "hist");              /* 12 */
+  dmap_add_int(playlists, "ceQi", -200);                   /* 12 */
+  dmap_add_int(playlists, "ceQm", 200);                    /* 12 */
+  dmap_add_string(playlists, "ceQl", "Previously Played"); /* 25 = 8 + 17 */
+
+  if (songlist)
+    {
+      dmap_add_container(playlists, "mlit", 36);
+      dmap_add_string(playlists, "ceQk", "curr");            /* 12 */
+      dmap_add_int(playlists, "ceQi", 0);                    /* 12 */
+      dmap_add_int(playlists, "ceQm", 1);                    /* 12 */
+
+      dmap_add_container(playlists, "mlit", 69);
+      dmap_add_string(playlists, "ceQk", "main");            /* 12 */
+      dmap_add_int(playlists, "ceQi", 1);                    /* 12 */
+      dmap_add_int(playlists, "ceQm", n);                    /* 12 */
+      dmap_add_string(playlists, "ceQl", "Up Next");         /* 15 = 8 + 7 */
+      dmap_add_string(playlists, "ceQh", "from Music");      /* 18 = 8 + 10 */
+
+      songlist_length = EVBUFFER_LENGTH(songlist);
+    }
+  else
+    songlist_length = 0;
+
+  /* Final construction of reply */
+  dmap_add_container(evbuf, "ceQR", 79 + EVBUFFER_LENGTH(playlists) + songlist_length); /* size of entire container */
+  dmap_add_int(evbuf, "mstt", 200);                                                     /* 12, dmap.status */
+  dmap_add_int(evbuf, "mtco", abs(span));                                               /* 12 */
+  dmap_add_int(evbuf, "mrco", n);                                                       /* 12 */
+  dmap_add_char(evbuf, "ceQu", 0);                                                      /*  9 */
+  dmap_add_container(evbuf, "mlcl", 8 + EVBUFFER_LENGTH(playlists) + songlist_length);  /*  8 */
+  dmap_add_container(evbuf, "ceQS", EVBUFFER_LENGTH(playlists));                        /*  8 */
+  ret = evbuffer_add_buffer(evbuf, playlists);
+  evbuffer_free(playlists);
+  if (ret < 0)
+    {
+      DPRINTF(E_LOG, L_DACP, "Could not add playlists to evbuffer for playqueue-contents\n");
+      if (songlist)
+	evbuffer_free(songlist);
+
+      dmap_send_error(req, "ceQR", "Out of memory");
+      return;
+    }
+
+  if (songlist)
+    {
+      ret = evbuffer_add_buffer(evbuf, songlist);
+      evbuffer_free(songlist);
+      if (ret < 0)
+	{
+	  DPRINTF(E_LOG, L_DACP, "Could not add songlist to evbuffer for playqueue-contents\n");
+
+	  dmap_send_error(req, "ceQR", "Out of memory");
+	  return;
+	}
+    } 
+  dmap_add_char(evbuf, "apsm", status.shuffle); /*  9, daap.playlistshufflemode - not part of mlcl container */
+  dmap_add_char(evbuf, "aprm", status.repeat);  /*  9, daap.playlistrepeatmode  - not part of mlcl container */
+
+  httpd_send_reply(req, HTTP_OK, "OK", evbuf);
+}
+
+static void
+dacp_reply_playqueueedit_clear(struct evhttp_request *req, struct evbuffer *evbuf, char **uri, struct evkeyvalq *query)
+{
+  const char *param;
+  int clear_hist;
+
+  clear_hist = 0;
+  param = evhttp_find_header(query, "mode");
+
+  /*
+   * The mode parameter contains the playlist to be cleared.
+   * If mode=0x68697374 (hex representation of the ascii string "hist") clear the history,
+   * otherwise the current playlist.
+   */
+  if (strcmp(param,"0x68697374") == 0)
+    clear_hist = 1;
+
+  player_queue_empty(clear_hist);
+
+  dmap_add_container(evbuf, "cacr", 24); /* 8 + len */
+  dmap_add_int(evbuf, "mstt", 200);      /* 12 */
+  dmap_add_int(evbuf, "miid", 0);        /* 12 */
+
+  httpd_send_reply(req, HTTP_OK, "OK", evbuf);
+}
+
+static void
+dacp_reply_playqueueedit_add(struct evhttp_request *req, struct evbuffer *evbuf, char **uri, struct evkeyvalq *query)
+{
+  //?command=add&query='dmap.itemid:156'&sort=album&mode=3&session-id=100
+  // -> mode=3: add to playqueue position 0 (play next)
+  //?command=add&query='dmap.itemid:158'&sort=album&mode=0&session-id=100
+  // -> mode=0: add to end of playqueue
+  //?command=add&query='dmap.itemid:306'&queuefilter=album:6525753023700533274&sort=album&mode=1&session-id=100
+  // -> mode 1: stop playblack, clear playqueue, add songs to playqueue
+
+  struct player_source *ps;
+  const char *editquery;
+  const char *queuefilter;
+  const char *querymodifier;
+  const char *sort;
+  const char *param;
+  char modifiedquery[32];
+  uint32_t idx;
+  int mode;
+  int plid;
+  int ret;
+  int quirkyquery;
+
+  param = evhttp_find_header(query, "mode");
+  if (param)
+    {
+      ret = safe_atoi32(param, &mode);
+      if (ret < 0)
+	{
+	  DPRINTF(E_LOG, L_DACP, "Invalid mode value in playqueue-edit request\n");
+
+	  dmap_send_error(req, "cacr", "Invalid request");
+	  return;
+	}
+    }
+
+  if ((mode == 1) || (mode == 2))
+    {
+      player_playback_stop();
+      player_queue_clear();
+    }
+
+  editquery = evhttp_find_header(query, "query");
+  if (editquery)
+    {
+      sort = evhttp_find_header(query, "sort");
+
+      // if sort param is missing and an album or artist is added to the queue, set sort to "album"
+      if (!sort && (strstr(editquery, "daap.songalbumid:") || strstr(editquery, "daap.songartistid:")))
+      {
+	sort = "album";
+      }
+
+      // only use queryfilter if mode is not equal 3 (play next) or 5 (add to up next)
+      queuefilter = (mode == 3 || mode == 5) ? NULL : evhttp_find_header(query, "queuefilter");
+
+      querymodifier = evhttp_find_header(query, "query-modifier");
+      if (!querymodifier || (strcmp(querymodifier, "containers") != 0))
+	{
+	  quirkyquery = (mode == 1) && strstr(editquery, "dmap.itemid:") && ((!queuefilter) || strstr(queuefilter, "(null)"));
+	  ret = player_queue_make_daap(&ps, editquery, queuefilter, sort, quirkyquery);
+	}
+      else
+	{
+	  // Modify the query: Take the id from the editquery and use it as a queuefilter playlist id
+	  ret = safe_atoi32(strchr(editquery, ':') + 1, &plid);
+	  if (ret < 0)
+	    {
+	      DPRINTF(E_LOG, L_DACP, "Invalid playlist id in request: %s\n", editquery);
+
+	      dmap_send_error(req, "cacr", "Invalid request");
+	      return;
+	    }
+	  
+	  snprintf(modifiedquery, sizeof(modifiedquery), "playlist:%d", plid);
+	  ret = player_queue_make_daap(&ps, NULL, modifiedquery, sort, 0);
+	}
+
+      if (ret < 0)
+	{
+	  DPRINTF(E_LOG, L_DACP, "Could not build song queue\n");
+
+	  dmap_send_error(req, "cacr", "Invalid request");
+	  return;
+	}
+
+      idx = ret;
+
+      if (mode == 3)
+      {
+        player_queue_add_next(ps);
+      }
+      else
+      {
+        player_queue_add(ps);
+      }
+    }
+  else
+    {
+      DPRINTF(E_LOG, L_DACP, "Could not add song queue, DACP query missing\n");
+
+      dmap_send_error(req, "cacr", "Invalid request");
+      return;
+    }
+
+  if (mode == 2)
+    {
+      player_shuffle_set(1);
+      idx = 0;
+    }
+
+  DPRINTF(E_DBG, L_DACP, "Song queue built, playback starting at index %" PRIu32 "\n", idx);
+  ret = player_playback_start(&idx);
+  if (ret < 0)
+    {
+      DPRINTF(E_LOG, L_DACP, "Could not start playback\n");
+
+      dmap_send_error(req, "cacr", "Playback failed to start");
+      return;
+    }
+
+  /* 204 No Content is the canonical reply */
+  evhttp_send_reply(req, HTTP_NOCONTENT, "No Content", evbuf);
+}
+
+static void
+dacp_reply_playqueueedit_move(struct evhttp_request *req, struct evbuffer *evbuf, char **uri, struct evkeyvalq *query)
+{
+  /*
+   * Handles the move command.
+   * Exampe request:
+   * playqueue-edit?command=move&edit-params='edit-param.move-pair:3,0'&session-id=100
+   *
+   * The 'edit-param.move-pair' param contains the index of the song in the playqueue to be moved (index 3 in the example)
+   * and the index of the song after which it should be inserted (index 0 in the exampe, the now playing song).
+   */
+  int ret;
+
+  const char *param;
+  int src;
+  int dst;
+
+  param = evhttp_find_header(query, "edit-params");
+  if (param)
+  {
+    ret = safe_atoi32(strchr(param, ':') + 1, &src);
+    if (ret < 0)
+    {
+      DPRINTF(E_LOG, L_DACP, "Invalid edit-params move-from value in playqueue-edit request\n");
+
+      dmap_send_error(req, "cacr", "Invalid request");
+      return;
+    }
+
+    ret = safe_atoi32(strchr(param, ',') + 1, &dst);
+    if (ret < 0)
+    {
+      DPRINTF(E_LOG, L_DACP, "Invalid edit-params move-to value in playqueue-edit request\n");
+
+      dmap_send_error(req, "cacr", "Invalid request");
+      return;
+    }
+
+    player_queue_move(src, dst);
+  }
+
+  /* 204 No Content is the canonical reply */
+  evhttp_send_reply(req, HTTP_NOCONTENT, "No Content", evbuf);
+}
+
+static void
+dacp_reply_playqueueedit_remove(struct evhttp_request *req, struct evbuffer *evbuf, char **uri, struct evkeyvalq *query)
+{
+  /*
+   * Handles the remove command.
+   * Exampe request (removes song at position 1 in the playqueue):
+   * ?command=remove&items=1&session-id=100
+   */
+  int ret;
+
+  const char *param;
+  int item_index;
+
+  param = evhttp_find_header(query, "items");
+  if (param)
+  {
+    ret = safe_atoi32(param, &item_index);
+    if (ret < 0)
+    {
+      DPRINTF(E_LOG, L_DACP, "Invalid edit-params remove item value in playqueue-edit request\n");
+
+      dmap_send_error(req, "cacr", "Invalid request");
+      return;
+    }
+
+    player_queue_remove(item_index);
+  }
+
+  /* 204 No Content is the canonical reply */
+  evhttp_send_reply(req, HTTP_NOCONTENT, "No Content", evbuf);
+}
+
+static void
+dacp_reply_playqueueedit(struct evhttp_request *req, struct evbuffer *evbuf, char **uri, struct evkeyvalq *query)
+{
+  struct daap_session *s;
+  const char *param;
+
+  /*  Variations of /ctrl-int/1/playqueue-edit and expected behaviour
+      User selected play (album or artist tab):
+	?command=add&query='...'&sort=album&mode=1&session-id=...
+	-> clear queue, play query results
+      User selected play (playlist):
+	?command=add&query='dmap.itemid:...'&query-modifier=containers&mode=1&session-id=...
+	-> clear queue, play playlist with the id specified by itemid
+      User selected track (album tab):
+	?command=add&query='dmap.itemid:...'&queuefilter=album:...&sort=album&mode=1&session-id=...
+	-> clear queue, play itemid and the rest of album
+      User selected track (artist tab):
+	?command=add&query='dmap.itemid:...'&queuefilter=artist:...&sort=album&mode=1&session-id=...
+	-> clear queue, play itemid and the rest of artist tracks
+      User selected track (song tab):
+	?command=add&query='dmap.itemid:...'&queuefilter=playlist:...&sort=name&mode=1&session-id=...
+	-> clear queue, play itemid and the rest of playlist
+      User selected track (playlist tab):
+	?command=add&query='dmap.containeritemid:...'&queuefilter=playlist:...&sort=physical&mode=1&session-id=...
+	-> clear queue, play containeritemid and the rest of playlist
+      User selected shuffle (artist tab):
+	?command=add&query='...'&sort=album&mode=2&session-id=...
+	-> clear queue, play shuffled query results
+      User selected add item to queue:
+	?command=add&query='...'&sort=album&mode=0&session-id=...
+	-> add query results to queue
+      User selected play next song (album tab)
+	?command=add&query='daap.songalbumid:...'&sort=album&mode=3&session-id=...
+	-> replace queue from after current song with query results
+      User selected track in queue:
+	?command=playnow&index=...&session-id=...
+	-> play index
+
+      And the quirky query from Remote - no sort and no queuefilter
+      User selected track (Audiobooks):
+	?command=add&query='dmap.itemid:...'&mode=1&session-id=...
+	-> clear queue, play itemid and the rest of album tracks
+
+	?command=move&edit-params='edit-param.move-pair:3,0'&session-id=100
+	-> move song from playqueue position 3 to be played after song at position 0
+	?command=remove&items=1&session-id=100
+  -> remove song on position 1 from the playqueue
+   */
+
+  s = daap_session_find(req, query, evbuf);
+  if (!s)
+    return;
+
+  param = evhttp_find_header(query, "command");
+  if (!param)
+    {
+      DPRINTF(E_LOG, L_DACP, "No command in playqueue-edit request\n");
+
+      dmap_send_error(req, "cmst", "Invalid request");
+      return;
+    }
+
+  if (strcmp(param, "clear") == 0)
+    dacp_reply_playqueueedit_clear(req, evbuf, uri, query);
+  else if (strcmp(param, "playnow") == 0)
+    dacp_reply_cue_play(req, evbuf, uri, query);
+  else if (strcmp(param, "add") == 0)
+    dacp_reply_playqueueedit_add(req, evbuf, uri, query);
+  else if (strcmp(param, "move") == 0)
+    dacp_reply_playqueueedit_move(req, evbuf, uri, query);
+  else if (strcmp(param, "remove") == 0)
+    dacp_reply_playqueueedit_remove(req, evbuf, uri, query);
+  else
+    {
+      DPRINTF(E_LOG, L_DACP, "Unknown playqueue-edit command %s\n", param);
+
+      dmap_send_error(req, "cmst", "Invalid request");
+      return;
+    }
+}
+
 static void
 dacp_reply_playstatusupdate(struct evhttp_request *req, struct evbuffer *evbuf, char **uri, struct evkeyvalq *query)
 {
   struct daap_session *s;
   struct dacp_update_request *ur;
+  struct evhttp_connection *evcon;
   const char *param;
   int reqd_rev;
   int ret;
@@ -1076,7 +1709,7 @@ dacp_reply_playstatusupdate(struct evhttp_request *req, struct evbuffer *evbuf,
       return;
     }
 
-  if (reqd_rev == 1)
+  if ((reqd_rev == 0) || (reqd_rev == 1))
     {
       ret = make_playstatusupdate(evbuf);
       if (ret < 0)
@@ -1105,7 +1738,9 @@ dacp_reply_playstatusupdate(struct evhttp_request *req, struct evbuffer *evbuf,
   /* If the connection fails before we have an update to push out
    * to the client, we need to know.
    */
-  evhttp_connection_set_closecb(req->evcon, update_fail_cb, ur);
+  evcon = evhttp_request_get_connection(req);
+  if (evcon)
+    evhttp_connection_set_closecb(evcon, update_fail_cb, ur);
 }
 
 static void
@@ -1113,6 +1748,7 @@ dacp_reply_nowplayingartwork(struct evhttp_request *req, struct evbuffer *evbuf,
 {
   char clen[32];
   struct daap_session *s;
+  struct evkeyvalq *headers;
   const char *param;
   char *ctype;
   uint32_t id;
@@ -1182,10 +1818,11 @@ dacp_reply_nowplayingartwork(struct evhttp_request *req, struct evbuffer *evbuf,
 	goto no_artwork;
     }
 
-  evhttp_remove_header(req->output_headers, "Content-Type");
-  evhttp_add_header(req->output_headers, "Content-Type", ctype);
+  headers = evhttp_request_get_output_headers(req);
+  evhttp_remove_header(headers, "Content-Type");
+  evhttp_add_header(headers, "Content-Type", ctype);
   snprintf(clen, sizeof(clen), "%ld", (long)EVBUFFER_LENGTH(evbuf));
-  evhttp_add_header(req->output_headers, "Content-Length", clen);
+  evhttp_add_header(headers, "Content-Length", clen);
 
   /* No gzip compression for artwork */
   evhttp_send_reply(req, HTTP_OK, "OK", evbuf);
@@ -1452,7 +2089,12 @@ dacp_reply_setspeakers(struct evhttp_request *req, struct evbuffer *evbuf, char
   do
     {
       param++;
-      ret = safe_hextou64(param, &ids[i]);
+      /* Hyperfine Remote for Android will send in decimal, others use hex */
+      if ((strlen(param) > 1) && (param[1] != 'x'))
+	ret = safe_atou64(param, &ids[i]);
+      else
+	ret = safe_hextou64(param, &ids[i]);
+
       if (ret < 0)
 	{
 	  DPRINTF(E_LOG, L_DACP, "Invalid speaker id in request: %s\n", param);
@@ -1460,7 +2102,10 @@ dacp_reply_setspeakers(struct evhttp_request *req, struct evbuffer *evbuf, char
 	  nspk--;
 	  continue;
 	}
-
+      else
+	{
+	  DPRINTF(E_DBG, L_DACP, "Speaker id converted with ret %d, param %s, dec val %llu.\n", ret, param, ids[i]);
+	}
       i++;
     }
   while ((param = strchr(param + 1, ',')));
@@ -1537,6 +2182,14 @@ static struct uri_map dacp_handlers[] =
       .regexp = "^/ctrl-int/[[:digit:]]+/playstatusupdate$",
       .handler = dacp_reply_playstatusupdate
     },
+    {
+      .regexp = "^/ctrl-int/[[:digit:]]+/playqueue-contents$",
+      .handler = dacp_reply_playqueuecontents
+    },
+    {
+      .regexp = "^/ctrl-int/[[:digit:]]+/playqueue-edit$",
+      .handler = dacp_reply_playqueueedit
+    },
     {
       .regexp = "^/ctrl-int/[[:digit:]]+/nowplayingartwork$",
       .handler = dacp_reply_nowplayingartwork
@@ -1572,6 +2225,7 @@ dacp_request(struct evhttp_request *req)
   char *uri_parts[7];
   struct evbuffer *evbuf;
   struct evkeyvalq query;
+  struct evkeyvalq *headers;
   int handler;
   int ret;
   int i;
@@ -1663,9 +2317,10 @@ dacp_request(struct evhttp_request *req)
 
   evhttp_parse_query(full_uri, &query);
 
-  evhttp_add_header(req->output_headers, "DAAP-Server", "forked-daapd/" VERSION);
+  headers = evhttp_request_get_output_headers(req);
+  evhttp_add_header(headers, "DAAP-Server", "forked-daapd/" VERSION);
   /* Content-Type for all DACP replies; can be overriden as needed */
-  evhttp_add_header(req->output_headers, "Content-Type", "application/x-dmap-tagged");
+  evhttp_add_header(headers, "Content-Type", "application/x-dmap-tagged");
 
   dacp_handlers[handler].handler(req, evbuf, uri_parts, &query);
 
@@ -1757,6 +2412,7 @@ void
 dacp_deinit(void)
 {
   struct dacp_update_request *ur;
+  struct evhttp_connection *evcon;
   int i;
 
   player_set_update_handler(NULL);
@@ -1768,10 +2424,11 @@ dacp_deinit(void)
     {
       update_requests = ur->next;
 
-      if (ur->req->evcon)
+      evcon = evhttp_request_get_connection(ur->req);
+      if (evcon)
 	{
-	  evhttp_connection_set_closecb(ur->req->evcon, NULL, NULL);
-	  evhttp_connection_free(ur->req->evcon);
+	  evhttp_connection_set_closecb(evcon, NULL, NULL);
+	  evhttp_connection_free(evcon);
 	}
 
       free(ur);
