@@ -67,6 +67,9 @@ GCRY_THREAD_OPTION_PTHREAD_IMPL;
 # include "ffmpeg_url_evbuffer.h"
 #endif
 
+#ifdef HAVE_SPOTIFY_H
+# include "spotify.h"
+#endif
 
 #define PIDFILE   STATEDIR "/run/" PACKAGE ".pid"
 
@@ -569,9 +572,6 @@ main(int argc, char **argv)
 
   DPRINTF(E_LOG, L_MAIN, "Forked Media Server Version %s taking off\n", VERSION);
 
-  /* Initialize ffmpeg */
-  avcodec_init();
-
   ret = av_lockmgr_register(ffmpeg_lockmgr);
   if (ret < 0)
     {
@@ -582,6 +582,9 @@ main(int argc, char **argv)
     }
 
   av_register_all();
+#if LIBAVFORMAT_VERSION_MAJOR >= 54 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVFORMAT_VERSION_MINOR >= 13)
+  avformat_network_init();
+#endif
   av_log_set_callback(logger_ffmpeg);
 #if LIBAVFORMAT_VERSION_MAJOR < 53
   register_ffmpeg_evbuffer_url_protocol();
@@ -678,6 +681,15 @@ main(int argc, char **argv)
       goto filescanner_fail;
     }
 
+#ifdef HAVE_SPOTIFY_H
+  /* Spawn Spotify thread */
+  ret = spotify_init();
+  if (ret < 0)
+    {
+      DPRINTF(E_INFO, L_MAIN, "Spotify thread not started\n");;
+    }
+#endif
+
   /* Spawn player thread */
   ret = player_init();
   if (ret != 0)
@@ -786,6 +798,10 @@ main(int argc, char **argv)
   player_deinit();
 
  player_fail:
+#ifdef HAVE_SPOTIFY_H
+  DPRINTF(E_LOG, L_MAIN, "Spotify deinit\n");
+  spotify_deinit();
+#endif
   DPRINTF(E_LOG, L_MAIN, "File scanner deinit\n");
   filescanner_deinit();
 
@@ -817,6 +833,9 @@ main(int argc, char **argv)
 
  signal_block_fail:
  gcrypt_init_fail:
+#if LIBAVFORMAT_VERSION_MAJOR >= 54 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVFORMAT_VERSION_MINOR >= 13)
+  avformat_network_deinit();
+#endif
   av_lockmgr_register(NULL);
 
  ffmpeg_init_fail:
