@@ -6,7 +6,8 @@
 
 #if defined(__linux__)
 /* AirTunes v2 packet interval in ns */
-# define AIRTUNES_V2_STREAM_PERIOD   7980000
+/* (352 samples/packet * 1e9 ns/s) / 44100 samples/s = 7981859 ns/packet */
+# define AIRTUNES_V2_STREAM_PERIOD 7981859
 #elif defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
 /* AirTunes v2 packet interval in ms */
 # define AIRTUNES_V2_STREAM_PERIOD   8
@@ -20,6 +21,9 @@
 #define STOB(s) ((s) * 4)
 #define BTOS(b) ((b) / 4)
 
+/* Maximum number of previously played songs that are remembered */
+#define MAX_HISTORY_COUNT 20
+
 enum play_status {
   PLAY_STOPPED = 2,
   PLAY_PAUSED  = 3,
@@ -32,6 +36,12 @@ enum repeat_mode {
   REPEAT_ALL  = 2,
 };
 
+enum source_type {
+  SOURCE_FFMPEG = 0,
+  SOURCE_SPOTIFY,
+  SOURCE_PIPE,
+};
+
 struct spk_flags {
   unsigned selected:1;
   unsigned has_password:1;
@@ -55,7 +65,39 @@ struct player_status {
 typedef void (*spk_enum_cb)(uint64_t id, const char *name, int relvol, struct spk_flags flags, void *arg);
 typedef void (*player_status_handler)(void);
 
-struct player_source;
+struct player_source
+{
+  uint32_t id;
+
+  enum source_type type;
+  int setup_done;
+
+  uint64_t stream_start;
+  uint64_t output_start;
+  uint64_t end;
+
+  struct transcode_ctx *ctx;
+
+  struct player_source *pl_next;
+  struct player_source *pl_prev;
+
+  struct player_source *shuffle_next;
+  struct player_source *shuffle_prev;
+
+  struct player_source *play_next;
+};
+
+struct player_history
+{
+  /* Buffer index of the oldest remembered song */
+  unsigned int start_index;
+
+  /* Count of song ids in the buffer */
+  unsigned int count;
+
+  /* Circular buffer of song ids previously played by forked-daapd */
+  uint32_t id[MAX_HISTORY_COUNT];
+};
 
 
 int
@@ -67,7 +109,6 @@ player_get_status(struct player_status *status);
 int
 player_now_playing(uint32_t *id);
 
-
 void
 player_speaker_enumerate(spk_enum_cb cb, void *arg);
 
@@ -108,22 +149,38 @@ player_repeat_set(enum repeat_mode mode);
 int
 player_shuffle_set(int enable);
 
-
-struct player_source *
-player_queue_make_daap(const char *query, const char *sort);
+int
+player_queue_make_daap(struct player_source **head, const char *query, const char *queuefilter, const char *sort, int quirk);
 
 struct player_source *
 player_queue_make_pl(int plid, uint32_t *id);
 
+struct player_source *
+player_queue_get(void);
+
 int
 player_queue_add(struct player_source *ps);
 
+int
+player_queue_add_next(struct player_source *ps);
+
+int
+player_queue_move(int ps_pos_from, int ps_pos_to);
+
+int
+player_queue_remove(int ps_pos_remove);
+
 void
 player_queue_clear(void);
 
+void
+player_queue_empty(int clear_hist);
+
 void
 player_queue_plid(uint32_t plid);
 
+struct player_history *
+player_history_get(void);
 
 void
 player_set_update_handler(player_status_handler handler);
