@@ -85,6 +85,9 @@
 #define RAOP_MD_DELAY_STARTUP      15360
 #define RAOP_MD_DELAY_SWITCH       (RAOP_MD_DELAY_STARTUP * 2)
 
+/* This is an arbitrary value which just needs to be kept in sync with the config */
+#define RAOP_CONFIG_MAX_VOLUME     11
+
 struct raop_v2_packet
 {
   uint8_t clear[AIRTUNES_V2_PKT_LEN];
@@ -851,7 +854,7 @@ raop_metadata_prepare(int id, uint64_t rtptime)
   ret = artwork_get_item_filename(dbmfi.path, 600, 600, ART_CAN_PNG | ART_CAN_JPEG, rmd->artwork);
   if (ret < 0)
     {
-      DPRINTF(E_LOG, L_RAOP, "Failed to retrieve artwork for '%s' (%d); no artwork will be sent\n", dbmfi.title, id);
+      DPRINTF(E_INFO, L_RAOP, "Failed to retrieve artwork for '%s' (%d); no artwork will be sent\n", dbmfi.title, id);
 
       evbuffer_free(rmd->artwork);
       rmd->artwork = NULL;
@@ -1188,11 +1191,12 @@ raop_check_cseq(struct raop_session *rs, struct evrtsp_request *req)
   int request_cseq;
 
   reply_cseq = raop_grab_cseq(req->input_headers);
+  /* AirFoil won't return cseq, so skip the check */
   if (reply_cseq < 0)
     {
-      DPRINTF(E_LOG, L_RAOP, "No CSeq in reply\n");
+      DPRINTF(E_INFO, L_RAOP, "No CSeq in reply, skipping check\n");
 
-      return -1;
+      return 0;
     }
 
   request_cseq = raop_grab_cseq(req->output_headers);
@@ -1225,6 +1229,15 @@ raop_make_sdp(struct raop_session *rs, struct evrtsp_request *req, char *address
     "a=fmtp:96 %d 0 16 40 10 14 2 255 0 0 44100\r\n"			\
     "a=rsaaeskey:%s\r\n"						\
     "a=aesiv:%s\r\n"
+#define SDP_PLD_FMT_NO_ENC						\
+  "v=0\r\n"								\
+    "o=iTunes %u 0 IN IP4 %s\r\n"					\
+    "s=iTunes\r\n"							\
+    "c=IN IP4 %s\r\n"							\
+    "t=0 0\r\n"								\
+    "m=audio 0 RTP/AVP 96\r\n"						\
+    "a=rtpmap:96 AppleLossless\r\n"					\
+    "a=fmtp:96 %d 0 16 40 10 14 2 255 0 0 44100\r\n"
 
   char *p;
   int ret;
@@ -1233,10 +1246,14 @@ raop_make_sdp(struct raop_session *rs, struct evrtsp_request *req, char *address
   if (p)
     *p = '\0';
 
-  /* Add SDP payload */
-  ret = evbuffer_add_printf(req->output_buffer, SDP_PLD_FMT,
-			    session_id, address, rs->address, AIRTUNES_V2_PACKET_SAMPLES,
-			    raop_aes_key_b64, raop_aes_iv_b64);
+  /* Add SDP payload - but don't add RSA/AES key/iv if no encryption - important for ATV3 update 6.0 */
+  if (rs->encrypt)
+    ret = evbuffer_add_printf(req->output_buffer, SDP_PLD_FMT,
+			      session_id, address, rs->address, AIRTUNES_V2_PACKET_SAMPLES,
+			      raop_aes_key_b64, raop_aes_iv_b64);
+  else
+    ret = evbuffer_add_printf(req->output_buffer, SDP_PLD_FMT_NO_ENC,
+			      session_id, address, rs->address, AIRTUNES_V2_PACKET_SAMPLES);
 
   if (p)
     *p = '%';
@@ -1251,6 +1268,7 @@ raop_make_sdp(struct raop_session *rs, struct evrtsp_request *req, char *address
   return 0;
 
 #undef SDP_PLD_FMT
+#undef SDP_PLD_FMT_NO_ENC
 }
 
 
@@ -1586,25 +1604,28 @@ raop_send_req_announce(struct raop_session *rs, evrtsp_req_cb cb)
 
   evrtsp_add_header(req->output_headers, "Content-Type", "application/sdp");
 
-  /* Challenge */
-  gcry_randomize(challenge, sizeof(challenge), GCRY_STRONG_RANDOM);
-  challenge_b64 = b64_encode(challenge, sizeof(challenge));
-  if (!challenge_b64)
+  /* Challenge - but only if session is encrypted (important for ATV3 after update 6.0) */
+  if (rs->encrypt)
     {
-      DPRINTF(E_LOG, L_RAOP, "Couldn't encode challenge\n");
+      gcry_randomize(challenge, sizeof(challenge), GCRY_STRONG_RANDOM);
+      challenge_b64 = b64_encode(challenge, sizeof(challenge));
+      if (!challenge_b64)
+	{
+	  DPRINTF(E_LOG, L_RAOP, "Couldn't encode challenge\n");
 
-      goto cleanup_req;
+	  goto cleanup_req;
+	}
+
+      /* Remove base64 padding */
+      ptr = strchr(challenge_b64, '=');
+      if (ptr)
+	*ptr = '\0';
+
+      evrtsp_add_header(req->output_headers, "Apple-Challenge", challenge_b64);
+
+      free(challenge_b64);
     }
 
-  /* Remove base64 padding */
-  ptr = strchr(challenge_b64, '=');
-  if (ptr)
-    *ptr = '\0';
-
-  evrtsp_add_header(req->output_headers, "Apple-Challenge", challenge_b64);
-
-  free(challenge_b64);
-
   ret = evrtsp_make_request(rs->ctrl, req, EVRTSP_REQ_ANNOUNCE, rs->session_url);
   if (ret < 0)
     {
@@ -1689,6 +1710,7 @@ raop_session_cleanup(struct raop_session *rs)
 {
   struct raop_session *s;
   struct raop_v2_packet *pkt;
+  struct raop_v2_packet *next_pkt;
 
   if (rs == sessions)
     sessions = sessions->next;
@@ -1708,8 +1730,13 @@ raop_session_cleanup(struct raop_session *rs)
   /* No more active sessions, free retransmit buffer */
   if (!sessions)
     {
-      for (pkt = pktbuf_head; pkt; pkt = pkt->next)
-	free(pkt);
+      pkt = pktbuf_head;
+      while (pkt)
+	{
+	  next_pkt = pkt->next;
+	  free(pkt);
+	  pkt = next_pkt;
+	}
 
       pktbuf_head = NULL;
       pktbuf_tail = NULL;
@@ -1824,25 +1851,33 @@ raop_session_make(struct raop_device *rd, int family, raop_status_cb cb)
   rs->server_fd = -1;
 
   rs->password = rd->password;
+  rs->wants_metadata = rd->wants_metadata;
 
   switch (rd->devtype)
     {
-      case RAOP_DEV_APEX_80211G:
+      case RAOP_DEV_APEX1_80211G:
 	rs->encrypt = 1;
 	rs->auth_quirk_itunes = 1;
-	rs->wants_metadata = 0;
 	break;
 
-      case RAOP_DEV_APEX_80211N:
+      case RAOP_DEV_APEX2_80211N:
 	rs->encrypt = 1;
 	rs->auth_quirk_itunes = 0;
-	rs->wants_metadata = 0;
+	break;
+
+      case RAOP_DEV_APEX3_80211N:
+	rs->encrypt = 0;
+	rs->auth_quirk_itunes = 0;
 	break;
 
       case RAOP_DEV_APPLETV:
 	rs->encrypt = 0;
 	rs->auth_quirk_itunes = 0;
-	rs->wants_metadata = 1;
+	break;
+
+      case RAOP_DEV_OTHER:
+	rs->encrypt = rd->encrypt;
+	rs->auth_quirk_itunes = 0;
 	break;
     }
 
@@ -2204,9 +2239,26 @@ raop_metadata_send(int id, uint64_t rtptime, uint64_t offset, int startup)
 
 /* Volume handling */
 static float
-raop_volume_convert(int volume)
+raop_volume_convert(int volume, char *name)
 {
   float raop_volume;
+  cfg_t *airplay;
+  int max_volume;
+
+  max_volume = RAOP_CONFIG_MAX_VOLUME;
+
+  airplay = cfg_gettsec(cfg, "airplay", name);
+  if (airplay)
+    max_volume = cfg_getint(airplay, "max_volume");
+
+  if ((max_volume < 1) || (max_volume > RAOP_CONFIG_MAX_VOLUME))
+    {
+      DPRINTF(E_LOG, L_RAOP, "Config has bad max_volume (%d) for device %s, using default instead\n", max_volume, name);
+
+      max_volume = RAOP_CONFIG_MAX_VOLUME;
+    }
+
+  DPRINTF(E_DBG, L_RAOP, "Setting max_volume for device %s to %d\n", name, max_volume);
 
   /* RAOP volume
    *  -144.0 is off
@@ -2215,7 +2267,7 @@ raop_volume_convert(int volume)
   if (volume == 0)
     raop_volume = -144.0;
   else
-    raop_volume = -30.0 + ((float)volume * 30.0) / 100.0;
+    raop_volume = -30.0 + ((float)max_volume * (float)volume * 30.0) / (100.0 * RAOP_CONFIG_MAX_VOLUME);
 
   return raop_volume;
 }
@@ -2235,7 +2287,7 @@ raop_set_volume_internal(struct raop_session *rs, int volume, evrtsp_req_cb cb)
       return -1;
     }
 
-  raop_volume = raop_volume_convert(volume);
+  raop_volume = raop_volume_convert(volume, rs->devname);
 
   /* Don't let locales get in the way here */
   ret = evbuffer_add_printf(evbuf, "volume: %d.%06d\r\n", (int)raop_volume, -(int)(1000000.0 * (raop_volume - (int)raop_volume)));
@@ -3232,9 +3284,9 @@ raop_v2_resend_range(struct raop_session *rs, uint16_t seqnum, uint16_t len)
   uint16_t distance;
 
   /* Check that seqnum is in the retransmit buffer */
-  if ((seqnum > pktbuf_head->seqnum) && (seqnum < pktbuf_tail->seqnum))
+  if ((seqnum > pktbuf_head->seqnum) || (seqnum < pktbuf_tail->seqnum))
     {
-      DPRINTF(E_LOG, L_RAOP, "RAOP device %s asking for seqnum %u; not in buffer (h %u t %u)\n", rs->devname, seqnum, pktbuf_head->seqnum, pktbuf_tail->seqnum);
+      DPRINTF(E_WARN, L_RAOP, "RAOP device %s asking for seqnum %u; not in buffer (h %u t %u)\n", rs->devname, seqnum, pktbuf_head->seqnum, pktbuf_tail->seqnum);
       return;
     }
 
@@ -3259,12 +3311,12 @@ raop_v2_resend_range(struct raop_session *rs, uint16_t seqnum, uint16_t len)
 
   if (pktbuf == pktbuf_head)
     {
-      while (seqnum != pktbuf->seqnum)
+      while (pktbuf && seqnum != pktbuf->seqnum)
 	pktbuf = pktbuf->next;
     }
   else
     {
-      while (seqnum != pktbuf->seqnum)
+      while (pktbuf && seqnum != pktbuf->seqnum)
 	pktbuf = pktbuf->prev;
     }
 
