@@ -38,19 +38,28 @@
 # include <sys/endian.h>
 #endif
 
-#include <event.h>
-#include "evhttp/evhttp.h"
-
 #include <libavcodec/avcodec.h>
 #include <libavformat/avformat.h>
+#include <libavutil/mathematics.h>
+#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
+# include <libavutil/opt.h>
+# include <libavresample/avresample.h>
+#endif
 
 #include "logger.h"
 #include "conffile.h"
 #include "db.h"
 #include "transcode.h"
 
+#ifdef HAVE_SPOTIFY_H
+# include "spotify.h"
+#endif
 
-#define XCODE_BUFFER_SIZE ((AVCODEC_MAX_AUDIO_FRAME_SIZE * 3) / 2)
+#if LIBAVCODEC_VERSION_MAJOR >= 56 || (LIBAVCODEC_VERSION_MAJOR == 55 && LIBAVCODEC_VERSION_MINOR >= 18)
+# define XCODE_BUFFER_SIZE ((192000 * 3) / 2)
+#else
+# define XCODE_BUFFER_SIZE ((AVCODEC_MAX_AUDIO_FRAME_SIZE * 3) / 2)
+#endif
 
 
 struct transcode_ctx {
@@ -64,9 +73,13 @@ struct transcode_ctx {
   int16_t *abuffer;
 
   /* Resampling */
-  int need_resample;
-  int input_size;
+#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
+  AVAudioResampleContext *resample_ctx;
+#else
   ReSampleContext *resample_ctx;
+  int input_size;
+#endif
+  int need_resample;
   int16_t *re_abuffer;
 
   off_t offset;
@@ -146,6 +159,15 @@ transcode(struct transcode_ctx *ctx, struct evbuffer *evbuf, int wanted)
 #if BYTE_ORDER == BIG_ENDIAN
   int i;
 #endif
+#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
+  AVFrame *frame = NULL;
+  int got_frame;
+  int out_linesize;
+  int out_samples;
+#elif LIBAVCODEC_VERSION_MAJOR >= 54 || (LIBAVCODEC_VERSION_MAJOR == 53 && LIBAVCODEC_VERSION_MINOR >= 35)
+  AVFrame *frame = NULL;
+  int got_frame;
+#endif
 
   processed = 0;
 
@@ -162,17 +184,50 @@ transcode(struct transcode_ctx *ctx, struct evbuffer *evbuf, int wanted)
       /* Decode data */
       while (ctx->apacket2.size > 0)
 	{
+#if LIBAVCODEC_VERSION_MAJOR >= 56 || (LIBAVCODEC_VERSION_MAJOR == 55 && LIBAVCODEC_VERSION_MINOR >= 29)
+	  got_frame = 0;
+
+	  if (!frame)
+	    {
+	      frame = av_frame_alloc();
+	      if (!frame)
+		{
+		  DPRINTF(E_LOG, L_XCODE, "Out of memory for decoded frame\n");
+
+		  return -1;
+		}
+	    }
+	  else
+            av_frame_unref(frame);
+
+	  used = avcodec_decode_audio4(ctx->acodec,
+				       frame, &got_frame,
+				       &ctx->apacket2);
+#elif LIBAVCODEC_VERSION_MAJOR >= 54 || (LIBAVCODEC_VERSION_MAJOR == 53 && LIBAVCODEC_VERSION_MINOR >= 35)
+	  got_frame = 0;
+
+	  if (!frame)
+	    {
+	      frame = avcodec_alloc_frame();
+	      if (!frame)
+		{
+		  DPRINTF(E_LOG, L_XCODE, "Out of memory for decoded frame\n");
+
+		  return -1;
+		}
+	    }
+	  else
+            avcodec_get_frame_defaults(frame);
+
+	  used = avcodec_decode_audio4(ctx->acodec,
+				       frame, &got_frame,
+				       &ctx->apacket2);
+#else
 	  buflen = XCODE_BUFFER_SIZE;
 
-#if LIBAVCODEC_VERSION_MAJOR >= 53 || (LIBAVCODEC_VERSION_MAJOR == 52 && LIBAVCODEC_VERSION_MINOR >= 32)
-	  /* FFmpeg 0.6 */
 	  used = avcodec_decode_audio3(ctx->acodec,
 				       ctx->abuffer, &buflen,
 				       &ctx->apacket2);
-#else
-	  used = avcodec_decode_audio2(ctx->acodec,
-				       ctx->abuffer, &buflen,
-				       ctx->apacket2.data, ctx->apacket2.size);
 #endif
 
 	  if (used < 0)
@@ -186,10 +241,68 @@ transcode(struct transcode_ctx *ctx, struct evbuffer *evbuf, int wanted)
 	  ctx->apacket2.size -= used;
 
 	  /* No frame decoded this time around */
+#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
+	  if (!got_frame)
+	    continue;
+#elif LIBAVCODEC_VERSION_MAJOR >= 54 || (LIBAVCODEC_VERSION_MAJOR == 53 && LIBAVCODEC_VERSION_MINOR >= 35)
+	  if (!got_frame)
+	    continue;
+	  else
+	    {
+	      /* This part is from the libav wrapper for avcodec_decode_audio3 */
+	      int ch, plane_size;
+	      int planar = av_sample_fmt_is_planar(ctx->acodec->sample_fmt);
+	      int data_size = av_samples_get_buffer_size(&plane_size, ctx->acodec->channels, 
+				frame->nb_samples, ctx->acodec->sample_fmt, 1);
+
+	      if (XCODE_BUFFER_SIZE < data_size)
+		{
+		  DPRINTF(E_WARN, L_XCODE, "Output buffer too small for frame (%d < %d)\n", XCODE_BUFFER_SIZE, data_size);
+
+		  continue;
+		}
+
+	      memcpy(ctx->abuffer, frame->extended_data[0], plane_size);
+
+	      if (planar && ctx->acodec->channels > 1)
+		{
+		  uint8_t *out = ((uint8_t *)ctx->abuffer) + plane_size;
+		  for (ch = 1; ch < ctx->acodec->channels; ch++)
+		    {
+		      memcpy(out, frame->extended_data[ch], plane_size);
+		      out += plane_size;
+		    }
+		}
+	      buflen = data_size;
+	    }
+#else
 	  if (buflen == 0)
 	    continue;
+#endif
 
 	  if (ctx->need_resample)
+#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
+	    {
+	      av_samples_get_buffer_size(&out_linesize, 2, frame->nb_samples, AV_SAMPLE_FMT_S16, 0);
+
+	      out_samples = avresample_convert(ctx->resample_ctx, (uint8_t **)&ctx->re_abuffer, out_linesize, frame->nb_samples,
+	                                       (uint8_t **)frame->data, frame->linesize[0], frame->nb_samples);
+	      if (out_samples < 0)
+		{
+		  DPRINTF(E_LOG, L_XCODE, "Resample returned no samples!\n");
+
+		  return -1;
+		}
+
+	      buflen = out_samples * 2 * 2; /* 16bit samples, 2 channels */
+	      buf = ctx->re_abuffer;
+	    }
+	  else
+	    {
+	      buf = (int16_t *)frame->data[0];
+	      buflen = av_samples_get_buffer_size(NULL, ctx->acodec->channels, frame->nb_samples, ctx->acodec->sample_fmt, 1);
+	    }
+#else
 	    {
 	      buflen = audio_resample(ctx->resample_ctx, ctx->re_abuffer, ctx->abuffer, buflen / ctx->input_size);
 
@@ -204,6 +317,7 @@ transcode(struct transcode_ctx *ctx, struct evbuffer *evbuf, int wanted)
 	    }
 	  else
 	    buf = ctx->abuffer;
+#endif
 
 #if BYTE_ORDER == BIG_ENDIAN
 	  /* swap buffer, LE16 */
@@ -247,6 +361,17 @@ transcode(struct transcode_ctx *ctx, struct evbuffer *evbuf, int wanted)
 
   ctx->offset += processed;
 
+#if LIBAVCODEC_VERSION_MAJOR >= 56 || (LIBAVCODEC_VERSION_MAJOR == 55 && LIBAVCODEC_VERSION_MINOR >= 29)
+  if (frame)
+    av_frame_free(&frame);
+#elif LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
+  if (frame)
+    avcodec_free_frame(&frame);
+#else
+  if (frame)
+    av_free(frame);
+#endif
+
   return processed;
 }
 
@@ -337,12 +462,10 @@ transcode_seek(struct transcode_ctx *ctx, int ms)
   return got_ms;
 }
 
-
-struct transcode_ctx *
-transcode_setup(struct media_file_info *mfi, off_t *est_size, int wavhdr)
+int
+transcode_setup(struct transcode_ctx **nctx, struct media_file_info *mfi, off_t *est_size, int wavhdr)
 {
   struct transcode_ctx *ctx;
-  int i;
   int ret;
 
   ctx = (struct transcode_ctx *)malloc(sizeof(struct transcode_ctx));
@@ -350,11 +473,11 @@ transcode_setup(struct media_file_info *mfi, off_t *est_size, int wavhdr)
     {
       DPRINTF(E_WARN, L_XCODE, "Could not allocate transcode context\n");
 
-      return NULL;
+      return -1;
     }
   memset(ctx, 0, sizeof(struct transcode_ctx));
 
-#if LIBAVFORMAT_VERSION_MAJOR >= 53 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVCODEC_VERSION_MINOR >= 3)
+#if LIBAVFORMAT_VERSION_MAJOR >= 54 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVFORMAT_VERSION_MINOR >= 3)
   ret = avformat_open_input(&ctx->fmtctx, mfi->path, NULL, NULL);
 #else
   ret = av_open_input_file(&ctx->fmtctx, mfi->path, NULL, 0, NULL);
@@ -364,10 +487,14 @@ transcode_setup(struct media_file_info *mfi, off_t *est_size, int wavhdr)
       DPRINTF(E_WARN, L_XCODE, "Could not open file %s: %s\n", mfi->fname, strerror(AVUNERROR(ret)));
 
       free(ctx);
-      return NULL;
+      return -1;
     }
 
+#if LIBAVFORMAT_VERSION_MAJOR >= 54 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVFORMAT_VERSION_MINOR >= 3)
+  ret = avformat_find_stream_info(ctx->fmtctx, NULL);
+#else
   ret = av_find_stream_info(ctx->fmtctx);
+#endif
   if (ret < 0)
     {
       DPRINTF(E_WARN, L_XCODE, "Could not find stream info: %s\n", strerror(AVUNERROR(ret)));
@@ -375,14 +502,21 @@ transcode_setup(struct media_file_info *mfi, off_t *est_size, int wavhdr)
       goto setup_fail;
     }
 
+#if LIBAVCODEC_VERSION_MAJOR >= 53 || (LIBAVCODEC_VERSION_MAJOR == 52 && LIBAVCODEC_VERSION_MINOR >= 64)
+  ctx->astream = av_find_best_stream(ctx->fmtctx, AVMEDIA_TYPE_AUDIO, -1, -1, &ctx->adecoder, 0);
+
+  if (ctx->astream < 0)
+    {
+      DPRINTF(E_WARN, L_XCODE, "Did not find audio stream or suitable decoder for %s\n", mfi->fname);
+
+      goto setup_fail;
+    }
+#else
+  int i;
   ctx->astream = -1;
   for (i = 0; i < ctx->fmtctx->nb_streams; i++)
     {
-#if LIBAVCODEC_VERSION_MAJOR >= 53 || (LIBAVCODEC_VERSION_MAJOR == 52 && LIBAVCODEC_VERSION_MINOR >= 64)
-      if (ctx->fmtctx->streams[i]->codec->codec_type == AVMEDIA_TYPE_AUDIO)
-#else
       if (ctx->fmtctx->streams[i]->codec->codec_type == CODEC_TYPE_AUDIO)
-#endif
 	{
 	  ctx->astream = i;
 
@@ -397,20 +531,27 @@ transcode_setup(struct media_file_info *mfi, off_t *est_size, int wavhdr)
       goto setup_fail;
     }
 
-  ctx->acodec = ctx->fmtctx->streams[ctx->astream]->codec;
-
-  ctx->adecoder = avcodec_find_decoder(ctx->acodec->codec_id);
+  ctx->adecoder = avcodec_find_decoder(ctx->fmtctx->streams[ctx->astream]->codec->codec_id);
   if (!ctx->adecoder)
     {
       DPRINTF(E_WARN, L_XCODE, "No suitable decoder found for codec\n");
 
       goto setup_fail;
     }
+#endif
+  ctx->acodec = ctx->fmtctx->streams[ctx->astream]->codec;
 
   if (ctx->adecoder->capabilities & CODEC_CAP_TRUNCATED)
     ctx->acodec->flags |= CODEC_FLAG_TRUNCATED;
 
+#if LIBAVCODEC_VERSION_MAJOR >= 54 || (LIBAVCODEC_VERSION_MAJOR == 53 && LIBAVCODEC_VERSION_MINOR >= 6)
+  ctx->acodec->request_sample_fmt = AV_SAMPLE_FMT_S16;
+  ctx->acodec->request_channel_layout = AV_CH_LAYOUT_STEREO;
+
+  ret = avcodec_open2(ctx->acodec, ctx->adecoder, NULL);
+#else
   ret = avcodec_open(ctx->acodec, ctx->adecoder);
+#endif
   if (ret != 0)
     {
       DPRINTF(E_WARN, L_XCODE, "Could not open codec: %s\n", strerror(AVUNERROR(ret)));
@@ -426,15 +567,63 @@ transcode_setup(struct media_file_info *mfi, off_t *est_size, int wavhdr)
       goto setup_fail_codec;
     }
 
-  if ((ctx->acodec->sample_fmt != SAMPLE_FMT_S16)
-      || (ctx->acodec->channels != 2)
-      || (ctx->acodec->sample_rate != 44100))
+  ctx->need_resample = (ctx->acodec->sample_fmt != AV_SAMPLE_FMT_S16)
+                       || (ctx->acodec->channels != 2)
+                       || (ctx->acodec->sample_rate != 44100);
+
+  if (ctx->need_resample)
     {
+#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
+      if (!ctx->acodec->channel_layout)
+	{
+	  DPRINTF(E_DBG, L_XCODE, "Resample requires channel_layout, but none from ffmpeg. Setting to default.\n");
+
+	  ctx->acodec->channel_layout = av_get_default_channel_layout(ctx->acodec->channels);
+	}
+
+      DPRINTF(E_DBG, L_XCODE, "Will resample, decoded stream is: %s, %d channels (layout %" PRIu64 "), %d Hz\n",
+              av_get_sample_fmt_name(ctx->acodec->sample_fmt),  ctx->acodec->channels,
+              ctx->acodec->channel_layout, ctx->acodec->sample_rate);
+
+      ctx->resample_ctx = avresample_alloc_context();
+      if (!ctx->resample_ctx)
+	{
+	  DPRINTF(E_LOG, L_XCODE, "Out of memory for resample context\n");
+
+	  goto setup_fail_codec;
+	}
+
+      av_opt_set_int(ctx->resample_ctx, "in_sample_fmt",      ctx->acodec->sample_fmt, 0);
+      av_opt_set_int(ctx->resample_ctx, "out_sample_fmt",     AV_SAMPLE_FMT_S16, 0);
+      av_opt_set_int(ctx->resample_ctx, "in_channel_layout",  ctx->acodec->channel_layout, 0);
+      av_opt_set_int(ctx->resample_ctx, "out_channel_layout", AV_CH_LAYOUT_STEREO, 0);
+      av_opt_set_int(ctx->resample_ctx, "in_sample_rate",     ctx->acodec->sample_rate, 0);
+      av_opt_set_int(ctx->resample_ctx, "out_sample_rate",    44100, 0);
+
+      ret = avresample_open(ctx->resample_ctx);
+      if (ret < 0)
+	{
+	  DPRINTF(E_LOG, L_XCODE, "Could not open resample context\n");
+
+	  avresample_free(&ctx->resample_ctx);
+	  goto setup_fail_codec;
+	}
+
+      ctx->re_abuffer = av_realloc(ctx->re_abuffer, XCODE_BUFFER_SIZE * 2);
+      if (!ctx->re_abuffer)
+	{
+	  DPRINTF(E_LOG, L_XCODE, "Could not allocate resample buffer\n");
+
+	  avresample_free(&ctx->resample_ctx);
+	  goto setup_fail_codec;
+	}
+
+#else
       DPRINTF(E_DBG, L_XCODE, "Setting up resampling (%d@%d)\n", ctx->acodec->channels, ctx->acodec->sample_rate);
 
       ctx->resample_ctx = av_audio_resample_init(2,              ctx->acodec->channels,
 						 44100,          ctx->acodec->sample_rate,
-						 SAMPLE_FMT_S16, ctx->acodec->sample_fmt,
+						 AV_SAMPLE_FMT_S16, ctx->acodec->sample_fmt,
 						 16, 10, 0, 0.8);
 
       if (!ctx->resample_ctx)
@@ -452,9 +641,11 @@ transcode_setup(struct media_file_info *mfi, off_t *est_size, int wavhdr)
 	  audio_resample_close(ctx->resample_ctx);
 	  goto setup_fail_codec;
 	}
+#endif
 
-      ctx->need_resample = 1;
-#if LIBAVUTIL_VERSION_MAJOR >= 51 || (LIBAVUTIL_VERSION_MAJOR == 51 && LIBAVUTIL_VERSION_MINOR >= 4)
+#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
+
+#elif LIBAVUTIL_VERSION_MAJOR >= 52 || (LIBAVUTIL_VERSION_MAJOR == 51 && LIBAVUTIL_VERSION_MINOR >= 4)
       ctx->input_size = ctx->acodec->channels * av_get_bytes_per_sample(ctx->acodec->sample_fmt);
 #elif LIBAVCODEC_VERSION_MAJOR >= 53
       ctx->input_size = ctx->acodec->channels * av_get_bits_per_sample_fmt(ctx->acodec->sample_fmt) / 8;
@@ -470,16 +661,22 @@ transcode_setup(struct media_file_info *mfi, off_t *est_size, int wavhdr)
   if (wavhdr)
     make_wav_header(ctx, est_size);
 
-  return ctx;
+  *nctx = ctx;
+
+  return 0;
 
  setup_fail_codec:
   avcodec_close(ctx->acodec);
 
  setup_fail:
+#if LIBAVFORMAT_VERSION_MAJOR >= 54 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVFORMAT_VERSION_MINOR >= 21)
+  avformat_close_input(&ctx->fmtctx);
+#else
   av_close_input_file(ctx->fmtctx);
+#endif
   free(ctx);
 
-  return NULL;
+  return -1;
 }
 
 void
@@ -488,14 +685,25 @@ transcode_cleanup(struct transcode_ctx *ctx)
   if (ctx->apacket.data)
     av_free_packet(&ctx->apacket);
 
-  avcodec_close(ctx->acodec);
-  av_close_input_file(ctx->fmtctx);
+  if (ctx->acodec)
+    avcodec_close(ctx->acodec);
+#if LIBAVFORMAT_VERSION_MAJOR >= 54 || (LIBAVFORMAT_VERSION_MAJOR == 53 && LIBAVFORMAT_VERSION_MINOR >= 21)
+  if (ctx->fmtctx)
+    avformat_close_input(&ctx->fmtctx);
+#else
+  if (ctx->fmtctx)
+    av_close_input_file(ctx->fmtctx);
+#endif
 
   av_free(ctx->abuffer);
 
   if (ctx->need_resample)
     {
+#if LIBAVCODEC_VERSION_MAJOR >= 55 || (LIBAVCODEC_VERSION_MAJOR == 54 && LIBAVCODEC_VERSION_MINOR >= 35)
+      avresample_free(&ctx->resample_ctx);
+#else
       audio_resample_close(ctx->resample_ctx);
+#endif
       av_free(ctx->re_abuffer);
     }
 
@@ -513,6 +721,12 @@ transcode_needed(struct evkeyvalq *headers, char *file_codectype)
   int size;
   int i;
 
+  if (!file_codectype)
+    {
+      DPRINTF(E_LOG, L_XCODE, "Can't proceed, codectype is unknown (null)\n");
+      return -1;
+    }
+
   DPRINTF(E_DBG, L_XCODE, "Determining transcoding status for codectype %s\n", file_codectype);
 
   lib = cfg_getsec(cfg, "library");
@@ -579,6 +793,12 @@ transcode_needed(struct evkeyvalq *headers, char *file_codectype)
 	    {
 	      DPRINTF(E_DBG, L_XCODE, "Client is Remote, using iTunes codecs\n");
 
+	      client_codecs = itunes_codecs;
+	    }
+	  else if (strncmp(user_agent, "AppleCoreMedia", strlen("AppleCoreMedia")) == 0)
+	    {
+	      DPRINTF(E_DBG, L_XCODE, "Client is AppleCoreMedia, using iTunes codecs\n");
+
 	      client_codecs = itunes_codecs;
 	    }
 	  else if (strncmp(user_agent, "Roku", strlen("Roku")) == 0)
