@@ -2,7 +2,12 @@
 #ifndef __TRANSCODE_H__
 #define __TRANSCODE_H__
 
-#include "evhttp/evhttp.h"
+#include <event.h>
+#ifdef HAVE_LIBEVENT2
+# include <event2/http.h>
+#else
+# include "evhttp/evhttp.h"
+#endif
 
 struct transcode_ctx;
 
@@ -12,8 +17,8 @@ transcode(struct transcode_ctx *ctx, struct evbuffer *evbuf, int wanted);
 int
 transcode_seek(struct transcode_ctx *ctx, int ms);
 
-struct transcode_ctx *
-transcode_setup(struct media_file_info *mfi, off_t *est_size, int wavhdr);
+int
+transcode_setup(struct transcode_ctx **nctx, struct media_file_info *mfi, off_t *est_size, int wavhdr);
 
 void
 transcode_cleanup(struct transcode_ctx *ctx);
